"""This module implements the CostFunctionGenerator base classes."""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING

from bqskit.qis.state.system import StateSystem

if TYPE_CHECKING:
    from bqskit.ir.opt.cost import CostFunction
    from bqskit.ir.circuit import Circuit
    from bqskit.qis.state.state import StateVector
    from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class CostFunctionGenerator(abc.ABC):
    """
    The CostFunctionGenerator base class.

    A CostFunctionGenerator in BQSKit is a function generator that can
    produce maps from circuit parameters to real numbers.

    The `gen_cost` method signature includes a circuit and target unitary
    or state as parameters. This allows a user to configure the generator
    and pass it to anything that does instantiation, like a synthesis pass,
    which in turn, will generate configured CostFunctions. This is useful
    since passes might be working with changing circuits and will need to
    regenerate CostFunctions from time to time.
    """

    @abc.abstractmethod
    def gen_cost(
        self,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector | StateSystem,
    ) -> CostFunction:
        """
        Generate a function from a circuit and target that maps params to cost.

        Args:
            circuit (Circuit): The circuit the cost function is generated for.

            target (UnitaryMatrix | StateVector | StateSystem): The target
                object.

        Returns:
            (CostFunction): The primitive cost function
                that can be directly passed to a minimizer. This maps
                parameters or circuit inputs to a cost.
        """

    def calc_cost(
        self,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector | StateSystem,
    ) -> float:
        """Generate and calculate the cost from the CostFunction."""
        return self.gen_cost(circuit, target).get_cost(circuit.params)

    def __call__(
        self,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector | StateSystem,
    ) -> float:
        """Generate and calculate the cost from the CostFunction."""
        return self.calc_cost(circuit, target)



"""This module implements the CycleInterval class."""
from __future__ import annotations

import logging
from typing import Any
from typing import Iterator
from typing import Tuple
from typing import Union

from typing_extensions import TypeGuard

from bqskit.utils.typing import is_integer
_logger = logging.getLogger(__name__)


class CycleInterval(Tuple[int, int]):
    """
    The CycleInterval class.

    Represents an inclusive range of cycles in a given circuit.
    """

    def __new__(
        cls,
        lower_or_tuple: int | tuple[int, int],
        upper: int | None = None,
    ) -> CycleInterval:
        """
        CycleInterval Constructor.

        Allows constructing a CycleInterval with either a tuple of ints
        or two ints.

        Args:
            lower_or_tuple (int | tuple[int, int]): Either the lower
                bound for the interval or the tuple of lower and upper
                bounds.

            upper (int | None): The upper bound for the interval. If a
                tuple is passed in for `lower_or_tuple` then this should
                be None.

        Returns:
            (CycleInterval): The constructed CycleInterval.

        Raises:
            ValueError: If `lower_or_tuple` is a tuple and 'upper' is
                not None, or if `lower_or_tuple` is an integer and
                `upper` is None.

            ValueError: If the lower bound is greater than the upper bound.

            ValueError: If either bound is negative.

        Notes:
            The lower and upper bounds are inclusive.
        """
        if upper is not None and not is_integer(upper):
            raise TypeError(
                f'Expected int or None for upper, got {type(upper)}.',
            )

        if isinstance(lower_or_tuple, tuple):
            if not CycleInterval.is_interval(lower_or_tuple):
                raise TypeError('Expected two integer arguments.')

            if upper is not None:
                raise ValueError('Unable to handle extra argument.')

            lower = lower_or_tuple[0]
            upper = lower_or_tuple[1]

        elif is_integer(lower_or_tuple) and is_integer(upper):
            lower = lower_or_tuple

        elif is_integer(lower_or_tuple) and upper is None:
            raise ValueError('Expected two integer arguments.')

        else:
            raise TypeError('Expected two integer arguments.')

        if lower > upper:
            raise ValueError(
                'Expected lower to be <= upper, got {lower} <= {upper}.',
            )

        if lower < 0:
            raise ValueError(
                'Expected positive integers, got {lower} and {upper}.',
            )

        return super().__new__(cls, (lower, upper))  # type: ignore

    @property
    def lower(self) -> int:
        """The interval's inclusive lower bound."""
        return self[0]

    @property
    def upper(self) -> int:
        """The interval's inclusive upper bound."""
        return self[1]

    @property
    def indices(self) -> list[int]:
        """The indices contained within the interval."""
        return list(range(self.lower, self.upper + 1))

    def __contains__(self, cycle_index: object) -> bool:
        """Return true if `cycle_index` is inside this interval."""
        return self.lower <= cycle_index <= self.upper  # type: ignore

    def __iter__(self) -> Iterator[int]:
        """Return an iterator for all indices contained in the interval."""
        return range(self.lower, self.upper + 1).__iter__()

    def __len__(self) -> int:
        """The length of the interval."""
        return self.upper - self.lower + 1

    def overlaps(self, other: IntervalLike) -> bool:
        """Return true if `other` overlaps with this interval."""
        if not CycleInterval.is_interval(other):
            raise TypeError(f'Expected CycleInterval, got {type(other)}.')

        other = CycleInterval(other)

        return self.lower <= other.upper and self.upper >= other.lower

    def intersection(self, other: IntervalLike) -> CycleInterval:
        """Return the range defined by both `self` and `other` interval."""
        if not CycleInterval.is_interval(other):
            raise TypeError(f'Expected CycleInterval, got {type(other)}.')

        other = CycleInterval(other)

        if not self.overlaps(other):
            raise ValueError('Empty intersection in interval.')

        return CycleInterval(
            max(self.lower, other.lower),
            min(self.upper, other.upper),
        )

    def union(self, other: IntervalLike) -> CycleInterval:
        """Return the range defined by `self` or `other` interval."""
        if not CycleInterval.is_interval(other):
            raise TypeError(f'Expected CycleInterval, got {type(other)}.')

        other = CycleInterval(other)

        if not self.overlaps(other) and (
            self.upper + 1 != other[0]
            and self.lower - 1 != other[1]
        ):
            raise ValueError('Union would lead to invalid interval.')

        return CycleInterval(
            min(self.lower, other.lower),
            max(self.upper, other.upper),
        )

    def __lt__(self, other: tuple[int, ...]) -> bool:
        """
        Return true if `self` comes before `other`.

        The less than operator defines a partial ordering.
        """
        if CycleInterval.is_interval(other):
            return self.upper < other[0]

        return NotImplemented

    @staticmethod
    def is_interval(interval: Any) -> TypeGuard[IntervalLike]:
        """Return true if `interval` is a IntervalLike."""
        if isinstance(interval, CycleInterval):
            return True

        if not isinstance(interval, tuple):
            _logger.log(0, 'Bounds is not a tuple.')
            return False

        if len(interval) != 2:
            _logger.log(
                0,
                'Expected interval to contain two values'
                f', got {len(interval)}.',
            )
            return False

        if not is_integer(interval[0]):
            _logger.log(
                0,
                'Expected integer values in interval'
                f', got {type(interval[0])}.',
            )
            return False

        if not is_integer(interval[1]):
            _logger.log(
                0,
                'Expected integer values in interval'
                f', got {type(interval[1])}.',
            )
            return False

        return True

    def __repr__(self) -> str:
        """Return a string representation of the interval."""
        return f'Interval(lower={self.lower}, upper={self.upper})'


IntervalLike = Union[Tuple[int, int], CycleInterval]



"""This module implements the ExtendBlockSizePass class."""
from __future__ import annotations

import logging
from typing import Sequence

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.circuitgate import CircuitGate
from bqskit.ir.operation import Operation
from bqskit.qis.graph import CouplingGraph
from bqskit.utils.typing import is_integer


_logger = logging.getLogger(__name__)


class ExtendBlockSizePass(BasePass):
    """Ensure all blocks are at least a given size."""

    def __init__(self, minimum_size: int | None = None) -> None:
        """
        Construct a ExtendBlockSizePass.

        Args:
            minimum_size (int | None): Extend all blocks to at least this
                size. If left as None, the minimum size will be the size of
                the smallest multi-qudit gate in the model.
        """
        if not is_integer(minimum_size) and minimum_size is not None:
            raise TypeError('Expected an integer or None for minimum size.')

        self.minimum_size = minimum_size

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        if circuit.num_qudits == 1:
            _logger.debug('Skipping extend pass for single-qudit circuit.')
            return

        minimum_size = self.minimum_size
        if minimum_size is None:
            minimum_size = min(
                g.num_qudits for g in data.gate_set
                if g.num_qudits != 1
            )

        if minimum_size is None:
            _logger.warning('No multi-qudit gates in gate set.')
            return

        if circuit.num_qudits < minimum_size:
            raise RuntimeError('Cannot extend block larger than circuit.')

        cg = data.connectivity

        # Find all small blocks
        small_blocks: list[tuple[int, int]] = []
        for cycle, op in circuit.operations_with_cycles():
            if isinstance(op.gate, CircuitGate):
                if op.gate.num_qudits < minimum_size:
                    small_blocks.append((cycle, op.location[0]))
        small_blocks.sort()

        # Select qubits to add
        qudits_to_extend: list[tuple[int, ...]] = []
        for block_point in small_blocks:
            cycle = block_point[0]
            op = circuit[block_point]

            num_to_add = minimum_size - op.gate.num_qudits
            qudits = list(op.location)
            added = []
            for _ in range(num_to_add):
                neighbors = self.get_neighbors(qudits, cg)

                if len(neighbors) == 0:
                    raise RuntimeError('Coupling graph is not fully connected.')

                if any(circuit.is_point_idle((cycle, n)) for n in neighbors):
                    for n in neighbors:
                        if circuit.is_point_idle((cycle, n)):
                            added.append(n)
                            qudits.append(n)
                            break
                else:
                    added.append(neighbors[0])
                    qudits.append(neighbors[0])
            qudits_to_extend.append(tuple(added))

        # Build extended operations
        new_ops: list[Operation] = []
        for block_point, to_add_qudits in zip(small_blocks, qudits_to_extend):
            op = circuit[block_point]
            op_circ: Circuit = op.gate._circuit.copy()  # type: ignore
            for qudit in to_add_qudits:
                op_circ.append_qudit(circuit.radixes[qudit])
            gate = CircuitGate(op_circ, True)
            location = tuple(op.location) + to_add_qudits
            new_op = Operation(gate, location, op.params)
            new_ops.append(new_op)

        # Extend operations
        circuit.batch_replace(small_blocks, new_ops)

    def get_neighbors(self, l: Sequence[int], cg: CouplingGraph) -> list[int]:
        """Return the neighbors of location `l` in `cg`."""
        neighbors = set()
        for q in l:
            neighbors.update(cg.get_neighbors_of(q))
        return list(neighbors.difference(l))



"""This module implements the various intermediate results classes."""
from __future__ import annotations

import logging
import pickle
from os import listdir
from os import mkdir
from os.path import exists
from re import findall

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.circuitgate import CircuitGate
from bqskit.ir.lang.qasm2.qasm2 import OPENQASM2Language
from bqskit.ir.operation import Operation
from bqskit.passes.util.converttou3 import ToU3Pass

_logger = logging.getLogger(__name__)


class SaveIntermediatePass(BasePass):
    """
    The SaveIntermediate class.

    The SaveIntermediatePass stores individual CircuitGates in pickle or qasm
    format.
    """

    def __init__(
        self,
        path_to_save_dir: str,
        project_name: str | None = None,
        save_as_qasm: bool = True,
    ) -> None:
        """
        Constructor for the SaveIntermediatePass.

        Args:
            path_to_save_dir (str): Path to the directory in which inter-
                qasm for circuit blocks should be saved.

            project_name (str): Name of the project files.

        Raises:
            ValueError: If `path_to_save_dir` is not an existing directory.
        """
        if exists(path_to_save_dir):
            self.pathdir = path_to_save_dir
            if self.pathdir[-1] != '/':
                self.pathdir += '/'
        else:
            raise ValueError(
                f'Path {path_to_save_dir} does not exist',
            )
        self.projname = project_name if project_name is not None \
            else 'unnamed_project'

        enum = 1
        if exists(self.pathdir + self.projname):
            while exists(self.pathdir + self.projname + f'_{enum}'):
                enum += 1
            self.projname += f'_{enum}'
            _logger.warning(
                f'Path {path_to_save_dir} already exists, '
                f'saving to {self.pathdir + self.projname} '
                'instead.',
            )

        mkdir(self.pathdir + self.projname)

        self.as_qasm = save_as_qasm

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see BasePass for more info."""

        # Gather and enumerate CircuitGates to save
        blocks_to_save: list[tuple[int, Operation]] = []
        for enum, op in enumerate(circuit):
            if isinstance(op.gate, CircuitGate):
                blocks_to_save.append((enum, op))

        # Set up path and file names
        structure_file = self.pathdir + self.projname + '/structure.pickle'
        block_skeleton = self.pathdir + self.projname + '/block_'
        num_digits = len(str(len(blocks_to_save)))

        structure_list: list[list[int]] = []
        # NOTE: Block numbers are gotten by iterating through the circuit so
        # there is no guarantee that the blocks were partitioned in that order.
        for enum, block in blocks_to_save:
            enum = str(enum).zfill(num_digits)  # type: ignore
            structure_list.append(block.location)  # type: ignore
            subcircuit = Circuit(block.num_qudits)
            subcircuit.append_gate(
                block.gate,
                list(
                    range(
                        block.num_qudits,
                    ),
                ),
                block.params,
            )
            subcircuit.unfold((0, 0))
            await ToU3Pass().run(subcircuit, PassData(subcircuit))
            if self.as_qasm:
                with open(block_skeleton + f'{enum}.qasm', 'w') as f:
                    f.write(OPENQASM2Language().encode(subcircuit))
            else:
                with open(
                    f'{block_skeleton}{enum}.pickle', 'wb',
                ) as f:
                    pickle.dump(subcircuit, f)

        with open(structure_file, 'wb') as f:
            pickle.dump(structure_list, f)


class RestoreIntermediatePass(BasePass):
    def __init__(self, project_directory: str, load_blocks: bool = True):
        """
        Constructor for the RestoreIntermediatePass.

        Args:
            project_directory (str): Path to the checkpoint block files. This
                directory must also contain a valid "structure.pickle" file.

            load_blocks (bool): If True, blocks in the project directory will
                be loaded to the block_list during the constructor. Otherwise
                the user must explicitly call load_blocks() themselves. Defaults
                to True.

        Raises:
            ValueError: If `project_directory` does not exist or if
                `structure.pickle` is invalid.
        """
        self.proj_dir = project_directory
        if not exists(self.proj_dir):
            raise TypeError(
                f"Project directory '{self.proj_dir}' does not exist.",
            )
        if not exists(self.proj_dir + '/structure.pickle'):
            raise TypeError(
                f'Project directory `{self.proj_dir}` does not '
                'contain `structure.pickle`.',
            )

        with open(self.proj_dir + '/structure.pickle', 'rb') as f:
            self.structure = pickle.load(f)

        if not isinstance(self.structure, list):
            raise TypeError('The provided `structure.pickle` is not a list.')

        self.block_list: list[str] = []
        if load_blocks:
            self.reload_blocks()

    def reload_blocks(self) -> None:
        """
        Updates the `block_list` variable with the current contents of the
        `proj_dir`.

        Raises:
            ValueError: if there are more block files than indices in the
            `structure.pickle`.
        """
        files = listdir(self.proj_dir)
        self.block_list = [f for f in files if 'block_' in f]
        if len(self.block_list) > len(self.structure):
            raise ValueError(
                'More block files than indicies in `structure.pickle`',
            )

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """
        Perform the pass's operation, see BasePass for more info.

        Raises:
            ValueError: if a block file and the corresponding index in
                `structure.pickle` are differnt lengths.
        """
        # If the circuit is empty, just append blocks in order
        if circuit.depth == 0:
            for block in self.block_list:
                # Get block
                block_num = int(findall(r'\d+', block)[0])
                with open(self.proj_dir + '/' + block) as f:
                    block_circ = OPENQASM2Language().decode(f.read())
                # Get location
                block_location = self.structure[block_num]
                if block_circ.num_qudits != len(block_location):
                    raise ValueError(
                        f'{block} and `structure.pickle` locations are '
                        'different sizes.',
                    )
                # Append to circuit
                circuit.append_circuit(block_circ, block_location)
        # Check if the circuit has been partitioned, if so, try to replace
        # blocks



"""This module implements the Language base class and LangException class."""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from bqskit.ir.circuit import Circuit


class LangException(Exception):
    """Exceptions related to language encoding and decoding."""


class Language(abc.ABC):
    """The Language base class."""

    @abc.abstractmethod
    def encode(self, circuit: Circuit) -> str:
        """Write `circuit` in this language."""

    @abc.abstractmethod
    def decode(self, source: str) -> Circuit:
        """Parse `source` into a circuit."""



"""This module implements the QPredictDecompositionPass class."""
from __future__ import annotations

import logging
from typing import Any
from typing import Sequence

import numpy as np

from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.parameterized.unitary import VariableUnitaryGate
from bqskit.ir.location import CircuitLocation
from bqskit.ir.opt.cost import CostFunctionGenerator
from bqskit.ir.opt.cost import HilbertSchmidtResidualsGenerator
from bqskit.passes.synthesis.synthesis import SynthesisPass
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.math import pauli_expansion
from bqskit.utils.math import unitary_log_no_i
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_real_number

_logger = logging.getLogger(__name__)


class QPredictDecompositionPass(SynthesisPass):
    """
    The QPredictDecompositionPass class.

    Breaks down the input unitary into blocks.
    """

    def __init__(
        self,
        block_size_start: int = 2,
        block_size_limit: int | None = None,
        fail_limit: int = 3,
        success_threshold: float = 1e-8,
        progress_threshold_r: float = 5e-2,
        progress_threshold_a: float = 1e-8,
        cost: CostFunctionGenerator = HilbertSchmidtResidualsGenerator(),
        max_depth: int | None = None,
        instantiate_options: dict[str, Any] = {},
    ) -> None:
        """
        QPredictDecompositionPass Constructor.

        Args:
            block_size_start (int): The smallest block size to append
                each step. (Default: 2)

            block_size_limit (int | None): The largest block size to append
                each step. If left as None, the unitary being synthesized
                provides the limit. (Default: None)

            fail_limit (int): The amount of tries to make progress before
                increasing block size. (Default: 2)

            success_threshold (float): The distance threshold that
                determines successful termintation. Measured in cost
                described by the cost function. (Default: 1e-6)

            progress_threshold (float): The distance necessary to improve
                for the synthesis algorithm to complete a layer and move
                on. Lowering this will led to synthesis going deeper quicker,
                and raising it will force to algorithm to spend more time
                on each layer. Caution, changing this too much might break
                the synthesis algorithm. (Default: 5e-3)

            cost (CostFunction | None): The cost function that determines
                distance during synthesis. The goal of this synthesis pass
                is to implement circuits for the given unitaries that have
                a cost less than the `success_threshold`.
                (Default: HSDistance())

            max_depth (int): The maximum number of gates to append without
                success before termination. If left as None it will default
                to unlimited. (Default: None)

            instantiate_options (dict[str: Any]): Options passed directly
                to circuit.instantiate when instantiating circuit
                templates. (Default: {})

        Raises:
            ValueError: If `block_size_start` is nonpositive.

            ValueError: If `block_size_limit` is less than `block_size_start`.

            ValueError: If `fail_limit` is nonpositive.

            ValueError: If `max_depth` is nonpositive.
        """

        if not is_integer(block_size_start):
            raise TypeError(
                'Expected block_size_start to be an integer, got %s'
                % type(block_size_start),
            )

        if block_size_start <= 0:
            raise ValueError(
                'Expected block_size_start to be positive, got %d.'
                % block_size_start,
            )

        if block_size_limit is not None and not is_integer(block_size_limit):
            raise TypeError(
                'Expected block_size_limit to be an integer, got %s'
                % type(block_size_limit),
            )

        if block_size_limit is not None and block_size_limit < block_size_start:
            raise ValueError(
                'Expected block_size_limit to be larger than block_size_start,'
                'got %d.' % block_size_limit,
            )

        if not is_integer(fail_limit):
            raise TypeError(
                'Expected fail_limit to be an integer, got %s'
                % type(fail_limit),
            )

        if fail_limit <= 0:
            raise ValueError(
                'Expected fail_limit to be positive, got %d.' % fail_limit,
            )

        if not is_real_number(success_threshold):
            raise TypeError(
                'Expected real number for success_threshold'
                ', got %s' % type(success_threshold),
            )

        if not is_real_number(progress_threshold_r):
            raise TypeError(
                'Expected real number for progress_threshold_r'
                ', got %s' % type(progress_threshold_r),
            )

        if not is_real_number(progress_threshold_a):
            raise TypeError(
                'Expected real number for progress_threshold_a'
                ', got %s' % type(progress_threshold_a),
            )

        if not isinstance(cost, CostFunctionGenerator):
            raise TypeError(
                'Expected cost to be a CostFunctionGenerator, got %s'
                % type(cost),
            )

        if max_depth is not None and not is_integer(max_depth):
            raise TypeError(
                'Expected max_depth to be an integer, got %s'
                % type(max_depth),
            )

        if max_depth is not None and max_depth <= 0:
            raise ValueError(
                'Expected max_depth to be positive, got %d.' % int(max_depth),
            )

        if not isinstance(instantiate_options, dict):
            raise TypeError(
                'Expected dictionary for instantiate_options, got %s.'
                % type(instantiate_options),
            )

        self.block_size_start = block_size_start
        self.block_size_limit = block_size_limit
        self.fail_limit = fail_limit
        self.success_threshold = success_threshold
        self.progress_threshold_r = progress_threshold_r
        self.progress_threshold_a = progress_threshold_a
        self.cost = cost
        self.max_depth = max_depth
        self.instantiate_options = {
            'min_iters': 25,
            'diff_tol_r': 1e-4,
        }
        self.instantiate_options.update(instantiate_options)

    async def synthesize(
        self,
        utry: UnitaryMatrix | StateVector | StateSystem,
        data: PassData,
    ) -> Circuit:
        """Synthesize `utry`, see :class:`SynthesisPass` for more."""
        if isinstance(utry, (StateVector, StateSystem)):
            raise RuntimeError('Unable to synthesize a state with qpredict.')

        instantiate_options = self.instantiate_options.copy()
        if data.seed is not None:
            instantiate_options['seed'] = data.seed

        # 0. Skip any unitaries too small for the configured block.
        if self.block_size_start >= utry.num_qudits:
            _logger.warning(
                'Skipping synthesis: block size is too large.',
            )
            return Circuit.from_unitary(utry)

        # 1. Create empty circuit with same size and radixes as `utry`.
        circuit = Circuit(utry.num_qudits, utry.radixes)

        # 2. Calculate block sizes
        block_size_end = utry.num_qudits - 1
        if self.block_size_limit is not None:
            block_size_end = self.block_size_limit

        # 3. Calculate relevant coupling_graphs
        model = self.get_model(utry, data)
        locations = [
            model.get_locations(i)
            for i in range(self.block_size_start, block_size_end + 1)
        ]

        # 3. Bottom-up synthesis: build circuit up one gate at a time
        layer = 1
        last_cost = 1.0
        last_loc = None

        while True:
            remainder = utry.dagger @ circuit.get_unitary()
            sorted_locations = self.analyze_remainder(remainder, locations)

            for loc in sorted_locations:

                # Never predict the previous location
                if loc == last_loc:
                    continue

                _logger.info(f'Trying next predicted location {loc}.')
                circuit.append_gate(VariableUnitaryGate(len(loc)), loc)
                circuit.instantiate(
                    utry,
                    **instantiate_options,  # type: ignore
                )
                cost = self.cost.calc_cost(circuit, utry)
                _logger.info(f'Instantiated; layers: {layer}, cost: {cost:e}.')

                if cost < self.success_threshold:
                    _logger.info(f'Circuit found with cost: {cost:e}.')
                    _logger.info('Successful synthesis.')
                    return circuit

                progress_threshold = self.progress_threshold_a
                progress_threshold += self.progress_threshold_r * np.abs(cost)
                if last_cost - cost >= progress_threshold:
                    _logger.info('Progress has been made, depth increasing.')
                    last_loc = loc
                    last_cost = cost
                    layer += 1
                    break

                _logger.info('Progress has not been made.')
                circuit.pop((-1, loc[0]))

    def analyze_remainder(
        self,
        R: UnitaryMatrix,
        locations: Sequence[Sequence[CircuitLocation]],
    ) -> list[CircuitLocation]:
        """
        Perform remainder analysis on `R` to sort `locations`.

        Args:
            R (UnitaryMatrix): The remainder to analyze.

            locations (Sequence[Sequence[CircuitLocation]]): List of locations
                grouped by block size.

        Returns:
            (list[CircuitLocation]): Sorted list of locations for next block
                based on remainder analysis.
        """
        _logger.info('Performing remainder analysis.')
        pauli_coefs = pauli_expansion(unitary_log_no_i(R))  # type: ignore

        locations_by_index: dict[int, set[CircuitLocation]] = {}
        for location_group in locations:
            for loc in location_group:
                for qudit_index in loc:
                    if qudit_index not in locations_by_index:
                        locations_by_index[qudit_index] = set()

                    locations_by_index[qudit_index].add(loc)

        location_scores_by_size: dict[int, dict[CircuitLocation, float]] = {}
        for location_group in locations:
            for loc in location_group:
                if len(loc) not in location_scores_by_size:
                    location_scores_by_size[len(loc)] = {}
                location_scores_by_size[len(loc)][loc] = 0.0

        for coef_idx, coef in enumerate(pauli_coefs):
            for qudit_index in self.decode_qubits(coef_idx):
                for loc in locations_by_index[qudit_index]:
                    location_scores_by_size[len(loc)][loc] += np.abs(coef)

        sorted_locations_by_size: dict[int, list[CircuitLocation]] = {
            size:
            sorted(
                list(location_scores.keys()),
                key=lambda x: location_scores[x],
                reverse=True,
            )
            for size, location_scores in location_scores_by_size.items()
        }

        sorted_locations_length: dict[int, int] = {
            size: len(sorted_locations)
            for size, sorted_locations in sorted_locations_by_size.items()
        }

        sorted_locations_by_sorted_size = sorted(
            sorted_locations_by_size.items(),
            key=lambda x: x[0],
        )

        sorted_locations: list[CircuitLocation] = []
        for size, locations in sorted_locations_by_sorted_size:
            for i in range(self.fail_limit):
                if i < sorted_locations_length[size]:
                    sorted_locations.append(locations[i])

        sorted_locations.extend(sorted_locations_by_sorted_size[-1][1])

        _logger.debug('Found order of locations:')
        _logger.debug(sorted_locations)
        return sorted_locations

    def decode_qubits(self, pauli_coef_index: int) -> list[int]:
        """Decode `pauli_coef_index` into qubit indices."""
        qubit_indices = []
        qudit_index = 0
        while pauli_coef_index > 0:
            if ((pauli_coef_index & 3) >> 1) | (pauli_coef_index & 1):
                qubit_indices.append(qudit_index)
            pauli_coef_index >>= 2
            qudit_index += 1
        return qubit_indices



from __future__ import annotations

from typing import cast
from typing import Sequence

from bqskit.compiler.basepass import BasePass
from bqskit.passes.alias import PassAlias
from bqskit.utils.typing import is_sequence


class PassGroup(PassAlias):
    """A pass that is a group of other passes."""

    def __init__(self, passes: BasePass | Sequence[BasePass]) -> None:
        """Group together one or more `passes`."""
        if not is_sequence(passes):
            passes = [cast(BasePass, passes)]

        if not isinstance(passes, list):
            passes = list(passes)

        self.passes: list[BasePass] = passes

        for p in self.passes:
            if not isinstance(p, BasePass):
                raise TypeError(f'Expected a Pass, got {type(p)}.')

    def get_passes(self) -> list[BasePass]:
        """Return the passes to be run, see :class:`PassAlias` for more."""
        return self.passes



"""This module implements the GeneralizedSabreRoutingPass class."""
from __future__ import annotations

import logging

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.passes.mapping.sabre import GeneralizedSabreAlgorithm


_logger = logging.getLogger(__name__)


class GeneralizedSabreRoutingPass(BasePass, GeneralizedSabreAlgorithm):
    """
    Uses the Sabre algorithm to route the circuit.

    See :class:`GeneralizedSabreAlgorithm` for more info.
    """

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        subgraph = data.connectivity
        if not subgraph.is_fully_connected():
            raise RuntimeError('Cannot route circuit on disconnected qudits.')

        pi = [i for i in range(circuit.num_qudits)]
        self.forward_pass(circuit, pi, subgraph, modify_circuit=True)
        data.final_mapping = [pi[x] for x in data.final_mapping]

        _logger.info(f'Finished routing with layout: {str(pi)}')



"""This module defines the ClusteringPartitioner pass."""
from __future__ import annotations

import logging

import numpy as np

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.utils.typing import is_integer

_logger = logging.getLogger(__name__)


class ClusteringPartitioner(BasePass):
    """
    The ClusteringPartitioner Pass.

    This pass forms partitions in the circuit by placing points in the circuit
    and clustering the gates around them.
    """

    def __init__(self, block_size: int = 3, num_points: int = 8) -> None:
        """
        Construct a ClusteringPartitioner.

        Args:
            block_size (int): Maximum size of partitioned blocks.
                (Default: 3)

            num_points (int): Total number of points to place and clusters
                to form. (Default: 8)

        Raises:
            ValueError: If `block_size` is less than 2.

            ValueError: if `num_points` is nonpositive.
        """

        if not is_integer(block_size):
            raise TypeError(
                f'Expected integer for block_size, got {type(block_size)}.',
            )

        if block_size < 2:
            raise ValueError(
                f'Expected block_size to be greater than 2, got {block_size}.',
            )

        if not is_integer(num_points):
            raise TypeError(
                f'Expected integer for num_points, got {type(num_points)}.',
            )

        if num_points < 1:
            raise ValueError(
                f'Expected num_points to positive, got {num_points}.',
            )

        self.block_size = block_size
        self.num_points = num_points

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""

        if self.block_size > circuit.num_qudits:
            _logger.warning(
                'Configured block size is greater than circuit size; '
                'blocking entire circuit.',
            )
            circuit.fold({
                qudit_index: (0, circuit.num_cycles)
                for qudit_index in range(circuit.num_qudits)
            })
            return

        for i in range(self.num_points):
            # Pick best region of 4 random points
            best_region = None
            best_gates = 0

            for _ in range(4):
                cycle = 0
                qudit = 0
                while True:
                    cycle = np.random.randint(circuit.num_cycles)
                    qudit = np.random.randint(circuit.num_qudits)
                    if not circuit.is_point_idle((cycle, qudit)):
                        break

                region = circuit.surround(
                    (cycle, qudit),
                    self.block_size,
                    fail_quickly=True,
                )
                num_gates = len(circuit[region])

                if num_gates > best_gates:
                    best_gates = num_gates
                    best_region = region

            if best_region is None:
                raise RuntimeError('Unable to find a region.')

            circuit.fold(best_region)



"""This module defines the QuickPartitioner pass."""
from __future__ import annotations

import logging
from typing import cast
from typing import Sequence

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.barrier import BarrierPlaceholder
from bqskit.ir.gates.circuitgate import CircuitGate
from bqskit.ir.location import CircuitLocation
from bqskit.ir.point import CircuitPoint
from bqskit.utils.typing import is_integer

_logger = logging.getLogger(__name__)


class QuickPartitioner(BasePass):
    """
    A partitioner that iterates over circuit gates only once.

    This pass forms partitions in the circuit by iterating over the operations
    in a topological order and binning them into blocks.
    """

    def __init__(self, block_size: int = 3) -> None:
        """
        Construct a QuickPartitioner.

        Args:
            block_size (int): Maximum size of partitioned blocks.
                (Default: 3)

        Raises:
            ValueError: If `block_size` is less than 2.
        """

        if not is_integer(block_size):
            raise TypeError(
                f'Expected integer for block_size, got {type(block_size)}.',
            )

        if block_size < 2:
            raise ValueError(
                f'Expected block_size to be greater than 2, got {block_size}.',
            )

        self.block_size = block_size

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """
        Partition gates in a circuit into a series of CircuitGates.

        Args:
            circuit (Circuit): Circuit to be partitioned.

            data (dict[str,Any]): Optional data unique to specific run.
        """
        # The partitioned circuit that will be built and returned
        partitioned_circuit = Circuit(circuit.num_qudits, circuit.radixes)

        # Tracks bins with at least one active_qudit
        active_bins: list[Bin | None] = [
            None for _ in range(circuit.num_qudits)
        ]

        # Tracks the first cycle in circuit not included in partitioned_circuit
        dividing_line: dict[int, int] = {
            i: 0 if circuit._front[i] is None else circuit._front[i].cycle  # type: ignore  # noqa
            for i in range(circuit.num_qudits)
        }

        # Inactive bins that cannot yet be put on partitioned_circuit
        pending_bins: list[Bin] = []

        # Track how many bins have been closed without processing
        num_closed = 0

        def close_bin_qudits(bin: Bin, loc: Sequence[int], cycle: int) -> bool:
            """Deactivate `qudits` in `bin`; return True if bin is inactive."""
            for q in loc:
                if q in bin.active_qudits:
                    bin.active_qudits.remove(q)
                    bin.ends[q] = cycle - 1

                if active_bins[q] == bin:
                    active_bins[q] = None

            # Check if the bin is completely inactive now
            if len(bin.active_qudits) == 0:
                pending_bins.append(bin)
                return True

            return False

        def process_pending_bins() -> None:
            """Add pending bins that can be added to the partitioned circuit."""
            need_to_reprocess = True
            while need_to_reprocess:
                need_to_reprocess = False
                to_remove = []

                for bin in pending_bins:
                    if all(
                        dividing_line[qudit] == start
                        for qudit, start in bin.starts.items()
                    ):
                        to_remove.append(bin)
                        subc = circuit.get_slice(bin.op_list)
                        loc = list(sorted(bin.qudits))

                        # Merge previously placed blocks if possible
                        merging = not isinstance(bin, BarrierBin)
                        while merging:
                            merging = False
                            for p in partitioned_circuit.rear:
                                op = partitioned_circuit[p]
                                if isinstance(op.gate, BarrierPlaceholder):
                                    # Don't merge through barriers
                                    continue
                                qudits = list(op.location)

                                # if qudits is subset of loc
                                if all(q in loc for q in qudits):
                                    prev_op = partitioned_circuit.pop(p)
                                    pg = cast(CircuitGate, prev_op.gate)
                                    prev_circ = pg._circuit.copy()
                                    local_loc = [loc.index(q) for q in qudits]
                                    subc.insert_circuit(0, prev_circ, local_loc)

                                    # retry merging
                                    merging = True
                                    break

                                # if loc is a subset of qudits
                                if all(q in qudits for q in loc):
                                    prev_op = partitioned_circuit.pop(p)
                                    pg = cast(CircuitGate, prev_op.gate)
                                    prev_circ = pg._circuit.copy()
                                    lloc = [qudits.index(q) for q in loc]
                                    prev_circ.append_circuit(subc, lloc)
                                    subc.become(prev_circ)
                                    loc = qudits

                                    # retry merging
                                    merging = True
                                    break

                        # Place circuit
                        partitioned_circuit.append_circuit(
                            subc,
                            loc,
                            not isinstance(bin, BarrierBin),
                            True,
                        )
                        for qudit in bin.qudits:
                            if bin.ends[qudit] is not None:
                                dividing_line[qudit] = bin.ends[qudit] + 1  # type: ignore # noqa
                            else:
                                dividing_line[qudit] = circuit.num_cycles

                        need_to_reprocess = True
                        break

                for bin in to_remove:
                    pending_bins.remove(bin)

        # Main loop
        for cycle, op in circuit.operations_with_cycles():
            point = CircuitPoint(cycle, op.location[0])
            location = op.location

            # Get all currently active bins that share at least one qudit
            overlapping_bins: list[Bin] = list({
                active_bins[q] for q in location  # type: ignore
                if active_bins[q] is not None
            })

            # Barriers close all overlapping bins
            if isinstance(op.gate, BarrierPlaceholder):
                for bin in overlapping_bins:
                    if close_bin_qudits(bin, location, cycle):
                        num_closed += 1
                    else:
                        extended = [q for q in location if q not in bin.qudits]
                        bin.blocked_qudits.update(extended)

                # Track the barrier to restore it in partitioned circuit
                pending_bins.append(BarrierBin(point, location, circuit))
                continue

            # Get all the currently active bins that can have op added to them
            admissible_bins = [
                bin for bin in overlapping_bins
                if bin.can_accommodate(location, self.block_size)
            ]

            # Close location on inadmissible overlapping bins
            for bin in overlapping_bins:
                if bin not in admissible_bins:
                    if close_bin_qudits(bin, location, cycle):
                        num_closed += 1

            # Select bin or make new one
            if len(admissible_bins) == 0:
                # If we cannot add this op to any bin, make a new one
                assert all(active_bins[q] is None for q in location)
                selected_bin = Bin()

            else:
                # Otherwise select an admissible bin
                found = False
                for bin in admissible_bins:
                    if all(q in bin.qudits for q in location):
                        selected_bin = bin
                        found = True
                        break

                if not found:
                    selected_bin = admissible_bins[0]

                # Close the overlapping qudits on the other admissible bins
                for bin in admissible_bins:
                    if bin != selected_bin:
                        if close_bin_qudits(bin, location, cycle):
                            num_closed += 1

            # Add op to selected_bin
            selected_bin.add_op(point, location)
            for q in location:
                if active_bins[q] is None:
                    active_bins[q] = selected_bin
                else:
                    assert active_bins[q] == selected_bin

            # Block qudits to prevent circular dependencies
            for active_bin in active_bins:
                if active_bin is None:
                    continue
                if active_bin == selected_bin:
                    continue

                indirect = active_bin.blocked_qudits
                indirect = indirect.union(active_bin.qudits)
                indirect = indirect.intersection(selected_bin.qudits)
                if len(indirect) != 0:
                    active_bin.blocked_qudits.update(selected_bin.qudits)
                    blockedqs = selected_bin.blocked_qudits
                    active_bin.blocked_qudits.update(blockedqs)

            # If a new bin was finalized, reprocess pending bins
            if num_closed >= 5:
                process_pending_bins()
                num_closed = 0

        # Close remaining active bins
        for b in active_bins:
            if b is not None:
                close_bin_qudits(b, b.qudits, circuit.num_cycles)

        # Process remaining bins
        process_pending_bins()

        if len(pending_bins) != 0:
            raise RuntimeError(
                'Unable to process all pending bins during partitioning.\n'
                'This should never happen and is a major issue'
                ', please make a bug report containing the input circuit.',
            )

        # Become partitioned circuit
        circuit.become(partitioned_circuit, False)


class Bin:
    """A Bin is where gates go as the QuickPartitioner sweeps a circuit."""

    id: int = 0
    """Unique ID counter for Bin instances."""

    def __init__(self) -> None:
        """Can start a new bin from an operation."""

        # The qudits in the bin
        self.qudits: list[int] = []

        # The starting cycles for each qudit (inclusive)
        self.starts: dict[int, int] = {}

        # The ending cycles for each qudit (inclusive)
        self.ends: dict[int, int | None] = {}

        # The qudits that can still accept new gates
        self.active_qudits: list[int] = []

        # Qudits that cannot be added to the bin
        self.blocked_qudits: set[int] = set()

        # Points for each operation in this bin
        self.op_list: list[CircuitPoint] = []

        self.id = Bin.id
        Bin.id += 1

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Bin) and self.id == other.id

    def __repr__(self) -> str:
        return 'Bin ' + str(self.id)

    def add_op(self, point: CircuitPoint, location: CircuitLocation) -> None:
        """Add an operation the bin."""
        for q in location:
            if q not in self.qudits:
                self.qudits.append(q)
                self.active_qudits.append(q)
                self.starts[q] = point.cycle
                self.ends[q] = None
        self.op_list.append(point)

    def can_accommodate(self, loc: CircuitLocation, block_size: int) -> bool:
        """
        Return true if the op can be added to this bin.

        An op can be added to the bin if all overlapping qudits are active in
        the bin and if the new bin won't be too large.
        """
        if any(
            q in self.blocked_qudits
            and q not in self.active_qudits
            for q in loc
        ):
            return False

        overlapping_qudits_are_active = all(
            q not in self.qudits or q in self.active_qudits
            for q in loc
        )

        size_limit = max(block_size, len(self.qudits))
        too_big = len(set(self.qudits + list(loc))) > size_limit

        return overlapping_qudits_are_active and not too_big


class BarrierBin(Bin):
    """A special bin made to mark and preserve barrier location."""

    def __init__(
        self,
        point: CircuitPoint,
        location: CircuitLocation,
        circuit: Circuit,
    ) -> None:
        """Initialize a BarrierBin with the point and location of a barrier."""
        super().__init__()

        # Add the barrier
        self.add_op(point, location)

        # Barriar bins fill the volume to the next gates

        nexts = circuit.next(point)
        ends: dict[int, int | None] = {q: None for q in location}
        for p in nexts:
            loc = circuit[p].location
            for q in loc:
                if q in ends and (ends[q] is None or ends[q] >= p.cycle):  # type: ignore # noqa # short-circuit safety for >=
                    ends[q] = p.cycle - 1

        self.ends = ends

        # Close the bin
        for q in location:
            self.active_qudits.remove(q)



"""This module implements the Minimization class."""
from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt

from bqskit.ir.gates.parameterized.unitary import VariableUnitaryGate
from bqskit.ir.opt.cost.functions import HilbertSchmidtResidualsGenerator
from bqskit.ir.opt.cost.generator import CostFunctionGenerator
from bqskit.ir.opt.instantiater import Instantiater
from bqskit.ir.opt.minimizer import Minimizer
from bqskit.ir.opt.minimizers.ceres import CeresMinimizer
from bqskit.ir.opt.multistartgens.random import RandomStartGenerator
from bqskit.qis.state.state import StateLike
from bqskit.qis.state.state import StateVector

if TYPE_CHECKING:
    from bqskit.ir.circuit import Circuit
    from bqskit.qis.state.system import StateSystem
    from bqskit.qis.state.system import StateSystemLike
    from bqskit.qis.unitary.unitarymatrix import UnitaryLike
    from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class Minimization(Instantiater):
    """The Minimization circuit instantiater."""

    def __init__(
        self,
        cost_fn_gen: CostFunctionGenerator = HilbertSchmidtResidualsGenerator(),
        minimizer: Minimizer = CeresMinimizer(),
        **kwargs: dict[str, Any],  # TODO: handle dist_tol and other options
    ) -> None:
        """
        Construct and configure a Minimization Instantiater.

        Args:
            cost_fn_gen (CostFunctionGenerator): This generator creates cost
                functions that are minimized.
                (Default: HilbertSchmidtGenerator())

            minimizer (Minimizer): The minimizer to use. If left as
                None, attempts to select best one.
        """

        if not isinstance(cost_fn_gen, CostFunctionGenerator):
            raise TypeError(
                'Expected CostFunctionGenerator, got %s.' % type(cost_fn_gen),
            )

        if not isinstance(minimizer, Minimizer):
            raise TypeError('Expected Minimizer, got %s.' % type(minimizer))

        self.cost_fn_gen = cost_fn_gen
        self.minimizer = minimizer

    def instantiate(
        self,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector | StateSystem,
        x0: npt.NDArray[np.float64],
    ) -> npt.NDArray[np.float64]:
        """Instantiate `circuit`, see Instantiater for more info."""
        cost = self.cost_fn_gen.gen_cost(circuit, target)
        return self.minimizer.minimize(cost, x0)

    @staticmethod
    def is_capable(circuit: Circuit) -> bool:
        """Return true if the circuit can be instantiated."""
        return all(
            not isinstance(g, VariableUnitaryGate)
            for g in circuit.gate_set
        )

    @staticmethod
    def get_violation_report(circuit: Circuit) -> str:
        """
        All circuits can be instantiated with minimization.

        See Instantiater for more info.
        """
        return (
            'Cannot instantiate a circuit with VariableUnitaryGates'
            ' via minimization.'
        )

    @staticmethod
    def get_method_name() -> str:
        """Return the name of this method."""
        return 'minimization'

    def multi_start_instantiate_inplace(
        self,
        circuit: Circuit,
        target: UnitaryLike | StateLike | StateSystemLike,
        num_starts: int,
    ) -> None:
        """
        Instantiate `circuit` to best implement `target` with multiple starts.

        See Instantiater for more info.
        """
        target = self.check_target(target)
        start_gen = RandomStartGenerator()
        starts = start_gen.gen_starting_points(num_starts, circuit, target)
        cost_fn = self.cost_fn_gen.gen_cost(circuit, target)
        params_list = [self.instantiate(circuit, target, x0) for x0 in starts]
        params = sorted(params_list, key=lambda x: cost_fn(x))[0]
        circuit.set_params(params)

    async def multi_start_instantiate_async(
        self,
        circuit: Circuit,
        target: UnitaryLike | StateLike | StateSystemLike,
        num_starts: int,
    ) -> Circuit:
        """
        Instantiate `circuit` to best implement `target` with multiple starts.

        See Instantiater for more info.
        """
        from bqskit.runtime import get_runtime
        target = self.check_target(target)
        start_gen = RandomStartGenerator()
        starts = start_gen.gen_starting_points(num_starts, circuit, target)
        cost_fn = self.cost_fn_gen.gen_cost(circuit, target)
        params_list = await get_runtime().map(
            self.instantiate,
            [circuit] * num_starts,
            [target] * num_starts,
            starts,
        )
        params = sorted(params_list, key=lambda x: cost_fn(x))[0]
        circuit.set_params(params)
        return circuit



"""This module implements the SycamoreGate."""
from __future__ import annotations

import cmath
import math

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class SycamoreGate(ConstantGate, QubitGate):
    """
    The SycamoreGate gate.

    The Sycamore gate is given by the following unitary:

    .. math::

        \\begin{pmatrix}
        1 & 0 & 0 & 0 \\\\
        0 & 0 & -i & 0 \\\\
        0 & -i & 0 & 0 \\\\
        0 & 0 & 0 & e^{-i\\frac{\\pi}{6}} \\\\
        \\end{pmatrix}
    """

    _num_qudits = 2
    _qasm_name = 'syc'
    _utry = UnitaryMatrix(
        [
            [1, 0, 0, 0],
            [0, 0, -1j, 0],
            [0, -1j, 0, 0],
            [0, 0, 0, cmath.exp(-1j * math.pi / 6)],
        ],
    )



"""This module implements the WideLayerGenerator class."""
from __future__ import annotations

import logging

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.gateset import GateSetLike
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gate import Gate
from bqskit.ir.gates.constant.itoffoli import IToffoliGate
from bqskit.ir.gates.parameterized.u3 import U3Gate
from bqskit.passes.search.generator import LayerGenerator
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
_logger = logging.getLogger(__name__)


class WideLayerGenerator(LayerGenerator):
    """
    A layer generator for use with wide gates like an iToffoli.

    Very similar to the simple layer generator. Each additional block adds one
    multi-qudit gate followed by single-qudit gates on each qudit.
    """

    def __init__(
        self,
        multi_qudit_gates: GateSetLike = IToffoliGate(),
        single_qudit_gate: Gate | None = U3Gate(),
    ) -> None:
        """
        Construct a WideLayerGenerator.

        Args:
            multi_qudit_gate (Gate | Sequence[Gate]): A multi-qudit gate
                or sequence of gates that starts this layer generator's
                building block. If multiple gates are given as a sequence,
                then a successive layer is built for each gate
                and each location. (Default: IToffoliGate())

            single_qudit_gate (Gate | None): A single-qudit gate that
                follows `multi_qudit_gate` in the building block. Can
                set to None to disable single-qudit gates, which might
                be useful with general multi-qudit gates.
                (Default: U3Gate())

        Raises:
            ValueError: If the single-qudit gate's size is not 1.

            ValueError: If there is any radix mismatch between the gates.
        """
        if isinstance(multi_qudit_gates, Gate):
            multi_qudit_gates = [multi_qudit_gates]

        if not all(isinstance(g, Gate) for g in multi_qudit_gates):
            raise TypeError(
                'Expected gate for multi_qudit_gates, got %s.'
                % [type(g) for g in multi_qudit_gates],
            )

        if single_qudit_gate is not None:
            if not isinstance(single_qudit_gate, Gate):
                raise TypeError(
                    'Expected gate for single_qudit_gate, got %s.'
                    % type(single_qudit_gate),
                )

            if single_qudit_gate.num_qudits != 1:
                raise ValueError(
                    'Expected single-qudit gate'
                    ', got a gate that acts on %d qudits.'
                    % single_qudit_gate.num_qudits,
                )

            sr = single_qudit_gate.radixes[0]
            for mg in multi_qudit_gates:
                if any(r != sr for r in mg.radixes):
                    raise ValueError(
                        'Radix mismatch between gates'
                        f': {mg.radixes} !~ {single_qudit_gate.radixes}.',
                    )

        self.multi_qudit_gates: list[Gate] = list(multi_qudit_gates)
        self.single_qudit_gate = single_qudit_gate

    def gen_initial_layer(
        self,
        target: UnitaryMatrix | StateVector | StateSystem,
        data: PassData,
    ) -> Circuit:
        """
        Generate the initial layer, see LayerGenerator for more.

        Raises:
            ValueError: If `target` has a radix mismatch with
                `self.initial_layer_gate`.
        """

        if not isinstance(target, (UnitaryMatrix, StateVector, StateSystem)):
            raise TypeError(
                'Expected unitary or state, got %s.' % type(target),
            )

        init_circuit = Circuit(target.num_qudits, target.radixes)

        if self.single_qudit_gate is None:
            return init_circuit

        for radix in target.radixes:
            if radix != self.single_qudit_gate.radixes[0]:
                raise ValueError(
                    'Radix mismatch between target and single_qudit_gate.',
                )

        for i in range(init_circuit.num_qudits):
            init_circuit.append_gate(self.single_qudit_gate, [i])

        return init_circuit

    def gen_successors(self, circuit: Circuit, data: PassData) -> list[Circuit]:
        """
        Generate the successors of a circuit node.

        Raises:
            ValueError: If circuit is a single-qudit circuit.
        """

        if not isinstance(circuit, Circuit):
            raise TypeError('Expected circuit, got %s.' % type(circuit))

        if circuit.num_qudits < 2:
            raise ValueError('Cannot expand a single-qudit circuit.')

        # Get the coupling graph
        cg = BasePass.get_connectivity(circuit, data)

        # Generate successors
        successors = []
        for mg in self.multi_qudit_gates:
            if mg.num_qudits > circuit.num_qudits:
                continue
            for loc in cg.get_subgraphs_of_size(mg.num_qudits):
                successor = circuit.copy()
                successor.append_gate(mg, loc)
                if self.single_qudit_gate is not None:
                    for q in loc:
                        successor.append_gate(self.single_qudit_gate, q)
                successors.append(successor)

        return successors



"""The Berkeley Quantum Synthesis Toolkit Python Package."""
from __future__ import annotations

import logging
from sys import stdout as _stdout

from .version import __version__  # noqa: F401
from .version import __version_info__  # noqa: F401
from bqskit.compiler.compile import compile
from bqskit.compiler.machine import MachineModel
from bqskit.ir.circuit import Circuit
from bqskit.ir.lang import register_language as _register_language
from bqskit.ir.lang.qasm2 import OPENQASM2Language as _qasm

# Initialize Logging
_logging_initialized = False


def enable_logging(verbose: bool = False) -> None:
    """
    Enable logging for BQSKit.

    Args:
        verbose (bool): If set to True, will print more verbose messages.
            Defaults to False.
    """
    global _logging_initialized
    if not _logging_initialized:
        _logger = logging.getLogger('bqskit')
        _handler = logging.StreamHandler(_stdout)
        _handler.setLevel(0)
        _fmt_header = '%(asctime)s.%(msecs)03d - %(levelname)-8s |'
        _fmt_message = ' %(name)s: %(message)s'
        _fmt = _fmt_header + _fmt_message
        _formatter = logging.Formatter(_fmt, '%H:%M:%S')
        _handler.setFormatter(_formatter)
        _logger.addHandler(_handler)
        _logging_initialized = True

    level = logging.DEBUG if verbose else logging.INFO
    logging.getLogger('bqskit').setLevel(level)


def disable_logging() -> None:
    """Disable logging for BQSKit."""
    logging.getLogger('bqskit').setLevel(logging.CRITICAL)


def enable_dashboard() -> None:
    import warnings
    warnings.warn(
        'Dask has been removed from BQSKit. As a result, the'
        ' enable_dashboard method has been removed.'
        'This warning will turn into an error in a future update.',
        DeprecationWarning,
    )


def disable_dashboard() -> None:
    import warnings
    warnings.warn(
        'Dask has been removed from BQSKit. As a result, the'
        ' disable_dashboard method has been removed.'
        'This warning will turn into an error in a future update.',
        DeprecationWarning,
    )


def disable_parallelism() -> None:
    import warnings
    warnings.warn(
        'The disable_parallelism method has been removed.'
        ' Instead, set the "num_workers" parameter to 1 during '
        'Compiler construction. This warning will turn into'
        'an error in a future update.',
        DeprecationWarning,
    )


def enable_parallelism() -> None:
    import warnings
    warnings.warn(
        'The enable_parallelism method has been removed.'
        ' Instead, set the "num_workers" parameter to 1 during '
        'Compiler construction. This warning will turn into'
        'an error in a future update.',
        DeprecationWarning,
    )


__all__ = [
    'compile',
    'MachineModel',
    'Circuit',
    'enable_logging',
    'disable_logging',
]

# Register supported languages
_register_language('qasm', _qasm())



from __future__ import annotations

from typing import cast
from typing import Sequence

import numpy as np

from bqskit.exec.results import RunnerResults
from bqskit.exec.runner import CircuitRunner
from bqskit.ir.circuit import Circuit


class SimulationRunner(CircuitRunner):
    """Simulate a circuit."""

    def run(self, circuit: Circuit) -> RunnerResults:
        """Execute the circuit, see CircuitRunner.run for more info."""
        state = np.reshape(circuit.get_unitary()[:, 0], (-1))
        probs = np.square(np.abs(state), dtype=np.float64)
        probs = cast(Sequence[float], probs)
        return RunnerResults(circuit.num_qudits, circuit.radixes, probs)



"""This module implements the FourParamGenerator class."""
from __future__ import annotations

import logging

from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.constant.cx import CNOTGate
from bqskit.ir.gates.parameterized.rx import RXGate
from bqskit.ir.gates.parameterized.ry import RYGate
from bqskit.ir.gates.parameterized.rz import RZGate
from bqskit.ir.gates.parameterized.u3 import U3Gate
from bqskit.passes.search.generator import LayerGenerator
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
_logger = logging.getLogger(__name__)


class FourParamGenerator(LayerGenerator):
    """
    The FourParamGenerator class.

    This is an optimized layer generator that uses commutativity rules
    to reduce the number of parameters per block. This also fixes the
    gate set to use cnots, ry, rz, and u3 gates. This is based on the
    following equivalences:

    U--C--U   U--C--Rz--Ry--Rz   U--C--Ry--Rz
       |    ~    |             ~    |
    U--X--U   U--X--Rx--Ry--Rx   U--X--Ry--Rx
    """

    def gen_initial_layer(
        self,
        target: UnitaryMatrix | StateVector | StateSystem,
        data: PassData,
    ) -> Circuit:
        """
        Generate the initial layer, see LayerGenerator for more.

        Raises:
            ValueError: If `target` is not qubit only.
        """

        if not isinstance(target, (UnitaryMatrix, StateVector, StateSystem)):
            raise TypeError(
                'Expected unitary or state, got %s.' % type(target),
            )

        if not target.is_qubit_only():
            raise ValueError('Cannot generate layers for non-qubit circuits.')

        init_circuit = Circuit(target.num_qudits, target.radixes)
        for i in range(init_circuit.num_qudits):
            init_circuit.append_gate(U3Gate(), [i])
        return init_circuit

    def gen_successors(self, circuit: Circuit, data: PassData) -> list[Circuit]:
        """
        Generate the successors of a circuit node.

        Raises:
            ValueError: If circuit is a single-qudit circuit.
        """

        if not isinstance(circuit, Circuit):
            raise TypeError('Expected circuit, got %s.' % type(circuit))

        if circuit.num_qudits < 2:
            raise ValueError('Cannot expand a single-qudit circuit.')

        # Get the machine model
        coupling_graph = data.connectivity

        # Generate successors
        successors = []
        for edge in coupling_graph:

            if self.count_outer_cnots(circuit, edge) >= 3:
                # No need to build circuits with more than 3 cnots in a row
                if circuit.num_qudits != 2:
                    # Guard on >2 qubit to prevent high-error glitches
                    continue

            successor = circuit.copy()
            successor.append_gate(CNOTGate(), edge)
            successor.append_gate(RYGate(), edge[0])
            successor.append_gate(RZGate(), edge[0])
            successor.append_gate(RYGate(), edge[1])
            successor.append_gate(RXGate(), edge[1])
            successors.append(successor)

        return successors

    def count_outer_cnots(self, circuit: Circuit, edge: tuple[int, int]) -> int:
        """
        Count how many uninterrupted 4-param cnot blocks are on `edge`.

        This will count backwards from the right-side of the circuit and stop
        when a cnot is encountered including other qudits.
        """
        rear_point = circuit._rear[edge[0]]
        num_cx_seen = 0

        if rear_point is None or rear_point.cycle != circuit.num_cycles - 1:
            return num_cx_seen

        while rear_point is not None:
            rear_points = circuit.prev(rear_point)

            if len(rear_points) == 0:
                break

            rear_point = rear_points.pop()

            if circuit[rear_point].num_qudits == 1:
                # Move past single-qubit gates
                continue

            cx_op = circuit[rear_point]

            if cx_op.location != edge:
                # If CX is on a different edge stop counting
                break

            num_cx_seen += 1

        return num_cx_seen



"""This package contains integrations with the PyTKet framework."""
from __future__ import annotations

from bqskit.ext.pytket.translate import bqskit_to_pytket
from bqskit.ext.pytket.translate import pytket_to_bqskit

__all__ = ['bqskit_to_pytket', 'pytket_to_bqskit']



"""This module implements the RuntimeMessage enum."""
from __future__ import annotations

from enum import IntEnum


class RuntimeMessage(IntEnum):
    """An message sent between processes in the BQSKit runtime."""
    CONNECT = 0
    DISCONNECT = 1
    STARTED = 2
    SHUTDOWN = 3
    ERROR = 4
    REQUEST = 5
    RESULT = 6
    SUBMIT = 7
    SUBMIT_BATCH = 8
    STATUS = 9
    LOG = 10
    CANCEL = 11
    WAITING = 12
    UPDATE = 13



"""This package contains integrations with the Cirq framework."""
from __future__ import annotations

from bqskit.ext.cirq.translate import bqskit_to_cirq
from bqskit.ext.cirq.translate import cirq_to_bqskit

__all__ = ['bqskit_to_cirq', 'cirq_to_bqskit']



"""This module implements the OPENQASM2Language class."""
from __future__ import annotations

from typing import TYPE_CHECKING

from bqskit.ir.lang.language import LangException
from bqskit.ir.lang.language import Language
from bqskit.ir.lang.qasm2.parser import parse
from bqskit.ir.lang.qasm2.visitor import OPENQASMVisitor

if TYPE_CHECKING:
    from bqskit.ir.circuit import Circuit


class OPENQASM2Language(Language):
    """The OPENQASM2Language class."""

    def encode(self, circuit: Circuit) -> str:
        """Write `circuit` in this language."""
        if not circuit.is_qubit_only():
            raise LangException('Only qubit circuits can be wrriten to qasm.')

        source = "OPENQASM 2.0;\ninclude \"qelib1.inc\";\n"
        source += f'qreg q[{circuit.num_qudits}];\n'
        for gate in circuit.gate_set:
            source += gate.get_qasm_gate_def()

        for op in circuit:
            source += op.get_qasm()

        return source

    def decode(self, source: str) -> Circuit:
        """Parse `source` into a circuit."""
        tree = parse(source)
        visitor = OPENQASMVisitor()
        visitor.visit_topdown(tree)
        return visitor.get_circuit()



"""This module implements the GroupSingleQuditGatePass."""
from __future__ import annotations

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.barrier import BarrierPlaceholder
from bqskit.ir.region import CircuitRegion


class GroupSingleQuditGatePass(BasePass):
    """
    The GroupSingleQuditGatePass Pass.

    This pass groups together consecutive single-qudit gates.
    """

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""

        # Go through each qudit individually
        for q in range(circuit.num_qudits):

            single_qubit_regions = []
            region_start = None

            for c in range(circuit.num_cycles):
                if circuit.is_point_idle((c, q)):
                    continue

                op = circuit[c, q]
                if (
                    op.num_qudits == 1
                    and not isinstance(op.gate, BarrierPlaceholder)
                ):
                    if region_start is None:
                        region_start = c
                else:
                    if region_start is not None:
                        region = CircuitRegion({q: (region_start, c - 1)})
                        single_qubit_regions.append(region)
                        region_start = None

            if region_start is not None:
                region = CircuitRegion(
                    {q: (region_start, circuit.num_cycles - 1)},
                )
                single_qubit_regions.append(region)
                region_start = None

            for region in reversed(single_qubit_regions):
                circuit.fold(region)



from __future__ import annotations

from typing import TYPE_CHECKING

from bqskit.exec.results import RunnerResults
from bqskit.exec.runner import CircuitRunner
from bqskit.ir.circuit import Circuit

if TYPE_CHECKING:
    from qiskit.providers.ibmq import IBMQBackend


class IBMQRunner(CircuitRunner):
    """Simulate a circuit."""

    def __init__(self, backend: IBMQBackend) -> None:
        """Setup an IBMQRunner to execute circuits on `backend`."""
        self.backend = backend

    def run(self, circuit: Circuit) -> RunnerResults:
        """Execute the circuit, see CircuitRunner.run for more info."""
        # 1. Check circuit and self.backend are compatible
        # TODO

        # 2. Convert to Qiskit IR
        from qiskit import QuantumCircuit
        qiskit_circ = QuantumCircuit.from_qasm_str(circuit.to('qasm'))
        qiskit_circ.measure_all()

        # 3. Run circuit
        result = self.backend.run(qiskit_circ).result()
        shots = result.results[0].shots
        probs = [0.0 for i in range(2 ** circuit.num_qudits)]
        for bit_str, count in result.get_counts().items():
            probs[int(bit_str, 2)] = count / shots

        return RunnerResults(circuit.num_qudits, circuit.radixes, probs)



"""This module implements the ParallelDo class."""
from __future__ import annotations

import logging
from typing import Callable
from typing import Iterable
from typing import TYPE_CHECKING

from bqskit.compiler.basepass import _sub_do_work
from bqskit.compiler.basepass import BasePass
from bqskit.compiler.workflow import Workflow
from bqskit.runtime import get_runtime
from bqskit.utils.typing import is_iterable

if TYPE_CHECKING:
    from bqskit.compiler.passdata import PassData
    from bqskit.compiler.workflow import WorkflowLike
    from bqskit.ir.circuit import Circuit


_logger = logging.getLogger(__name__)


class ParallelDo(BasePass):
    """
    The ParallelDo class.

    This is a control pass that executes a sequence of workflows in parallel.
    The branch that is accepted can either be the first to complete or one
    selected by a provided ordering.
    """

    def __init__(
        self,
        pass_sequences: Iterable[WorkflowLike],
        less_than: Callable[[Circuit, Circuit], bool],
        pick_fisrt: bool = False,
    ) -> None:
        """
        Construct a ParallelDo.

        Args:
            pass_sequences (Iterable[WorkflowLike]): The group of workflows
                to run in parallel.

            less_than (Callable[[Circuit, Circuit], bool]): Return True
                if the first circuit is preferred to the second one.
                This will be used to determine which output circuit to
                select.

            pick_first (bool): If true, then the pass will complete as
                soon as one of the workflows finishes and will return
                the first result. Defaults to False.
        """
        if not is_iterable(pass_sequences):
            bad_type = type(pass_sequences)
            raise TypeError(f'Expected sequence of workflows, got {bad_type}.')

        if not callable(less_than):
            bad_type = type(less_than)
            msg = f'Expected callable function for less_than, got {bad_type}'
            raise TypeError(msg)

        self.workflows = [Workflow(p) for p in pass_sequences]
        self.less_than = less_than
        self.pick_first = pick_fisrt

        if len(self.workflows) == 0:
            raise ValueError('Must specify at least one workflow.')

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        _logger.debug('Running pass sequences in parallel.')

        # Submit jobs to the runtime
        runtime = get_runtime()
        future = runtime.map(
            _sub_do_work,
            self.workflows,
            circuit=circuit,
            data=data,
        )

        # Wait for results
        if self.pick_first:
            circuits_and_ids = await runtime.next(future)  # Wake on next result
            circuits = [x[1] for x in circuits_and_ids]
            runtime.cancel(future)  # Cancel remaining
        else:
            circuits = await future

        # Find the best result
        best_circ, best_data = circuits[0]
        for _circ, _data in circuits[1:]:
            if self.less_than(_circ, best_circ):
                best_circ = _circ
                best_data = _data

        # Become best result
        circuit.become(best_circ)
        data.become(best_data)



"""This package defines passes and objects that control pass execution flow."""
from __future__ import annotations

from bqskit.passes.control.dothendecide import DoThenDecide
from bqskit.passes.control.dowhileloop import DoWhileLoopPass
from bqskit.passes.control.foreach import ClearAllBlockData
from bqskit.passes.control.foreach import ForEachBlockPass
from bqskit.passes.control.ifthenelse import IfThenElsePass
from bqskit.passes.control.paralleldo import ParallelDo
from bqskit.passes.control.predicate import PassPredicate
from bqskit.passes.control.predicates.change import ChangePredicate
from bqskit.passes.control.predicates.count import GateCountPredicate
from bqskit.passes.control.predicates.many import ManyQuditGatesPredicate
from bqskit.passes.control.predicates.multi import MultiPhysicalPredicate
from bqskit.passes.control.predicates.notpredicate import NotPredicate
from bqskit.passes.control.predicates.physical import PhysicalPredicate
from bqskit.passes.control.predicates.single import AllConstantSingleQuditGates
from bqskit.passes.control.predicates.single import HasGeneralSingleQuditGate
from bqskit.passes.control.predicates.single import NoSingleQuditGatesInModel
from bqskit.passes.control.predicates.single import SinglePhysicalPredicate
from bqskit.passes.control.predicates.single import ZXGatePredicate
from bqskit.passes.control.predicates.width import WidthPredicate
from bqskit.passes.control.whileloop import WhileLoopPass

__all__ = [
    'DoWhileLoopPass',
    'ForEachBlockPass',
    'ClearAllBlockData',
    'IfThenElsePass',
    'PassPredicate',
    'ChangePredicate',
    'GateCountPredicate',
    'ManyQuditGatesPredicate',
    'NotPredicate',
    'WhileLoopPass',
    'DoThenDecide',
    'ParallelDo',
    'WidthPredicate',
    'PhysicalPredicate',
    'MultiPhysicalPredicate',
    'SinglePhysicalPredicate',
    'NoSingleQuditGatesInModel',
    'HasGeneralSingleQuditGate',
    'ZXGatePredicate',
    'AllConstantSingleQuditGates',
]



"""This package contains instantiation implementations."""
# from bqskitrs import CeresInstantiater
from __future__ import annotations

from bqskit.ir.opt.instantiaters.minimization import Minimization
from bqskit.ir.opt.instantiaters.qfactor import QFactor


instantiater_order = [Minimization, QFactor]


__all__ = [
    'instantiater_order',
    'QFactor',
    'Minimization',
]



"""This module implements the GreedyHeuristic class."""
from __future__ import annotations

from bqskit.ir.circuit import Circuit
from bqskit.ir.opt.cost import CostFunctionGenerator
from bqskit.ir.opt.cost import HilbertSchmidtCostGenerator
from bqskit.passes.search.heuristic import HeuristicFunction
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class GreedyHeuristic(HeuristicFunction):
    """
    The GreedyHeuristic HeuristicFunction class.

    Defines a heuristic that results in greedy search. This function only looks
    at the current distance of the circuit from the target. This will create a
    behavior similar to depth-first search.
    """

    def __init__(
        self,
        cost_gen: CostFunctionGenerator = HilbertSchmidtCostGenerator(),
    ) -> None:
        """
        Construct a GreedyHeuristic Function.

        Args:
            cost_gen (CostFunctionGenerator): This is used to generate
                cost functions used during evaluations.
        """
        if not isinstance(cost_gen, CostFunctionGenerator):
            raise TypeError(
                'Expected CostFunctionGenerator for cost_gen, got %s.'
                % type(cost_gen),
            )

        self.cost_gen = cost_gen

    def get_value(
        self,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector | StateSystem,
    ) -> float:
        """Return the heuristic's value, see HeuristicFunction for more info."""
        return self.cost_gen.calc_cost(circuit, target)



"""This package implements various IO related passes."""
from __future__ import annotations

from bqskit.passes.io.checkpoint import LoadCheckpointPass
from bqskit.passes.io.checkpoint import SaveCheckpointPass
from bqskit.passes.io.intermediate import RestoreIntermediatePass
from bqskit.passes.io.intermediate import SaveIntermediatePass

__all__ = [
    'LoadCheckpointPass',
    'SaveCheckpointPass',
    'SaveIntermediatePass',
    'RestoreIntermediatePass',
]



"""
This subpackage implements Routing algorithms.

Routing algorithms are responsible routing logical gates via physical gates.
"""
from __future__ import annotations

from bqskit.passes.mapping.routing.pam import PAMRoutingPass
from bqskit.passes.mapping.routing.sabre import GeneralizedSabreRoutingPass

__all__ = ['GeneralizedSabreRoutingPass', 'PAMRoutingPass']



"""This module implements the IToffoliGate."""
from __future__ import annotations

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class IToffoliGate(ConstantGate, QubitGate):
    """
    The IToffoliGate gate, equal to an iX gate with two controls.

    The iToffoli gate is given by the following unitary:

    .. math::

        \\begin{pmatrix}
        1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\\\
        0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\\\
        0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\\\
        0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\\\
        0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\\\
        0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\\\
        0 & 0 & 0 & 0 & 0 & 0 & 0 & i \\\\
        0 & 0 & 0 & 0 & 0 & 0 & i & 0 \\\\
        \\end{pmatrix}

    References:
        Kim, Y., Morvan, A., Nguyen, L.B. et al. High-fidelity three-qubit
        iToffoli gate for fixed-frequency superconducting qubits. Nat. Phys.
        (2022). https://doi.org/10.1038/s41567-022-01590-3
    """

    _num_qudits = 3
    _qasm_name = 'iccx'
    _utry = UnitaryMatrix(
        [
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1j],
            [0, 0, 0, 0, 0, 0, 1j, 0],
        ],
    )



"""This module implements the IterativeScanningGateRemovalPass."""
from __future__ import annotations

import logging
from typing import Any

from bqskit.compiler.basepass import BasePass
from bqskit.passes.alias import PassAlias
from bqskit.passes.control import ChangePredicate
from bqskit.passes.control import ForEachBlockPass
from bqskit.passes.control import IfThenElsePass
from bqskit.passes.control import WhileLoopPass
from bqskit.passes.control import WidthPredicate
from bqskit.passes.partitioning import ScanPartitioner
from bqskit.passes.processing.scan import ScanningGateRemovalPass
from bqskit.passes.util import UnfoldPass
from bqskit.utils.typing import is_integer
_logger = logging.getLogger(__name__)


class IterativeScanningGateRemovalPass(PassAlias):
    """
    The IterativeScanningGateRemovalPass class.

    Starting from one side of the circuit, attempt to remove gates one-by-one.
    Repeat this until the circuit no longer changes. This will partition the
    circuit first if the circuit width is too large.
    """

    def __init__(
        self,
        width_to_partition: int = 5,
        block_size: int = 3,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Construct a IterativeScanningGateRemovalPass.

        Args:
            width_to_partition (int): Circuits at least as wide as this
                will be partitioned first.

            block_size (int): If the circuit is partitioned, it will
                be partitioned into blocks of this width.

            *args (Any): Arguments passed directly to
                :class:`ScanningGateRemovalPass`.

            *kwargs (Any): Keyword arguments passed directly to
                :class:`ScanningGateRemovalPass`.

        Raises:
            ValueError: If `width_to_partition` or `block_size` is
                nonpositive.

            ValueError: If `width_to_partition` is less than or equal to
                `block_size`.
        """

        if not is_integer(width_to_partition):
            raise TypeError(f'Expected an int, got {type(width_to_partition)}.')

        if not is_integer(block_size):
            raise TypeError(f'Expected an int, got {type(block_size)}.')

        if width_to_partition <= 0:
            raise ValueError(
                f'Expected positive width, got {width_to_partition}.',
            )

        if block_size <= 0:
            raise ValueError(f'Expected positive size, got {block_size}.')

        if width_to_partition <= block_size:
            raise ValueError(
                'Expected width to partition at to be greater'
                ' than block size.',
            )

        scan = ScanningGateRemovalPass(*args, **kwargs)
        self.passes: list[BasePass] = [
            WhileLoopPass(
                ChangePredicate(),
                [
                    IfThenElsePass(
                        WidthPredicate(width_to_partition),
                        scan,
                        [
                            ScanPartitioner(block_size),
                            ForEachBlockPass(scan),
                            UnfoldPass(),
                        ],
                    ),
                ],
            ),
        ]

    def get_passes(self) -> list[BasePass]:
        """Return the passes to be run, see :class:`PassAlias` for more."""
        return self.passes



"""This package contains LayerGenerator definitions."""
from __future__ import annotations

from bqskit.passes.search.generators.fourparam import FourParamGenerator
from bqskit.passes.search.generators.middleout import MiddleOutLayerGenerator
from bqskit.passes.search.generators.seed import SeedLayerGenerator
from bqskit.passes.search.generators.simple import SimpleLayerGenerator
from bqskit.passes.search.generators.single import SingleQuditLayerGenerator
from bqskit.passes.search.generators.stair import StairLayerGenerator
from bqskit.passes.search.generators.wide import WideLayerGenerator

__all__ = [
    'FourParamGenerator',
    'MiddleOutLayerGenerator',
    'SeedLayerGenerator',
    'SimpleLayerGenerator',
    'SingleQuditLayerGenerator',
    'StairLayerGenerator',
    'WideLayerGenerator',
]



"""This package contains several CircuitRunner implementations."""
from __future__ import annotations

__all__ = ['QuestRunner', 'SimulationRunner', 'gen_approximate_circuits']

try:
    from bqskit.exec.runners.ibmq import IBMQRunner
    __all__ += ['IBMQRunner']
except ImportError:
    pass

from bqskit.exec.runners.quest import QuestRunner
from bqskit.exec.runners.quest import gen_approximate_circuits
from bqskit.exec.runners.sim import SimulationRunner



"""This module implements the GeneralSQDecomposition."""
from __future__ import annotations

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.generalgate import GeneralGate


class GeneralSQDecomposition(BasePass):
    """Convert to a general single-qudit gate from the model's gate set."""

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""

        if circuit.num_qudits != 1:
            raise ValueError(
                'Cannot convert multi-qudit circuit into'
                ' general single-qudit gate.',
            )

        # Check radixes
        radix = circuit.radixes[0]

        # Pick a general gate
        general_gate: GeneralGate | None = None

        for gate in data.gate_set.single_qudit_gates:
            if isinstance(gate, GeneralGate):
                if gate.radixes[0] == radix:
                    general_gate = gate
                    break

        if general_gate is None:
            raise ValueError(
                f'No general single-qudit gate with radix {radix} in gate set.',
            )

        utry = circuit.get_unitary()
        new_circuit = Circuit(1)
        new_circuit.append_gate(general_gate, 0, general_gate.calc_params(utry))
        circuit.become(new_circuit)



"""This module implements the SGate."""
from __future__ import annotations

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class SGate(ConstantGate, QubitGate):
    """
    The single-qubit S gate.

    .. math::

        \\begin{pmatrix}
        1 & 0 \\\\
        0 & i \\\\
        \\end{pmatrix}
    """

    _num_qudits = 1
    _qasm_name = 's'
    _utry = UnitaryMatrix(
        [
            [1, 0],
            [0, 1j],
        ],
    )



"""This module implements the PhysicalPredicate class."""
from __future__ import annotations

from typing import TYPE_CHECKING

from bqskit.passes.control.predicate import PassPredicate

if TYPE_CHECKING:
    from bqskit.compiler.passdata import PassData
    from bqskit.ir.circuit import Circuit


class PhysicalPredicate(PassPredicate):
    """
    The PhysicalPredicate class.

    The PhysicalPredicate returns true if the circuit can be executed on the
    workflow's machine model with the current placement.
    """

    def get_truth_value(self, circuit: Circuit, data: PassData) -> bool:
        """Call this predicate, see :class:`PassPredicate` for more info."""
        return data.model.is_compatible(circuit, data.placement)



"""This module implements the HilbertSchmidtCost and HilbertSchmidtGenerator."""
from __future__ import annotations

from typing import TYPE_CHECKING

from bqskitrs import HilbertSchmidtResidualsFunction

from bqskit.ir.opt.cost.differentiable import DifferentiableResidualsFunction
from bqskit.ir.opt.cost.generator import CostFunctionGenerator
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem

if TYPE_CHECKING:
    from bqskit.ir.circuit import Circuit
    from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
    from bqskit.ir.opt.cost.function import CostFunction


class HilbertSchmidtResiduals(
    HilbertSchmidtResidualsFunction,
    DifferentiableResidualsFunction,
):
    """
    The HilbertSchmidtResiduals CostFunction implementation.

    The Hilbert-Schmidt Residuals CostFuction is a differentiable map from
    circuit parameters to a residual vector that is based on the Hilbert-Schmidt
    inner product. This function is global-phase-aware, meaning that the cost is
    zero if the target and circuit unitary differ only by a global phase.
    """


class HilbertSchmidtResidualsGenerator(CostFunctionGenerator):
    """
    The HilbertSchmidtResidualsGenerator class.

    This generator produces configured HilbertSchmidtResiduals functions.
    """

    def gen_cost(
        self,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector | StateSystem,
    ) -> CostFunction:
        """Generate a CostFunction, see CostFunctionGenerator for more info."""
        return HilbertSchmidtResiduals(circuit, target)



"""This module implements the TaggedGate Class."""
from __future__ import annotations

from typing import Any

import numpy as np
import numpy.typing as npt

from bqskit.ir.gate import Gate
from bqskit.ir.gates.composedgate import ComposedGate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.optimizable import LocallyOptimizableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.docs import building_docs


class TaggedGate(
    ComposedGate,
    LocallyOptimizableUnitary,
    DifferentiableUnitary,
):
    """
    The TaggedGate Class.

    Allows a user to place a tag on a gate.
    """

    def __init__(self, gate: Gate, tag: Any) -> None:
        """
        Associate `tag` with `gate`.

        Args:
            gate (Gate): The gate to tag.

            tag (Any): The tag to associate with the gate.
        """

        if not isinstance(gate, Gate):
            raise TypeError('Expected gate object, got %s' % type(gate))

        self.gate = gate
        self.tag = tag
        self._name = f'Tagged({gate.name}:{str(tag)})'
        self._num_params = gate.num_params
        self._num_qudits = gate.num_qudits
        self._radixes = gate.radixes

        # If input is a constant gate, we can cache the unitary.
        if self.num_params == 0 and not building_docs():
            self.utry = gate.get_unitary()

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        if hasattr(self, 'utry'):
            return self.utry

        return self.gate.get_unitary(params)

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        if hasattr(self, 'utry'):
            return np.array([])

        return self.gate.get_grad(params)  # type: ignore

    def get_unitary_and_grad(
        self,
        params: RealVector = [],
    ) -> tuple[UnitaryMatrix, npt.NDArray[np.complex128]]:
        """
        Return the unitary and gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        if hasattr(self, 'utry'):
            return self.utry, np.array([])

        return self.gate.get_unitary_and_grad(params)  # type: ignore

    def optimize(self, env_matrix: npt.NDArray[np.complex128]) -> list[float]:
        """
        Return the optimal parameters with respect to an environment matrix.

        See :class:`LocallyOptimizableUnitary` for more info.
        """
        if hasattr(self, 'utry'):
            return []

        self.check_env_matrix(env_matrix)
        return self.gate.optimize(env_matrix)  # type: ignore

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, TaggedGate)
            and self.gate == other.gate
            and self.tag == other.tag
        )

    def __hash__(self) -> int:
        if isinstance(self.tag, dict):
            return hash((self.gate, tuple(self.tag.items())))

        return hash((self.gate, self.tag))



"""This module implements the PermutationGate."""
from __future__ import annotations

from typing import Sequence

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.permutation import PermutationMatrix


class PermutationGate(ConstantGate, QubitGate):
    """A Permutation Gate."""

    def __init__(
        self,
        num_qudits: int,
        location: Sequence[int],
    ) -> None:
        """
        Construct a gate that shifts the state of qudits around.

        See :func:PermutationMatrix.from_qubit_location for more.

        Args:
            num_qubits (int): Total number of qubits

            location (Sequence[int]): The desired locations to swap
                the starting qubits to.

        Raises:
            ValueError: If num_qudits is nonpositive.
        """
        if num_qudits <= 0:
            raise ValueError('Expected positive integer, got %d' % num_qudits)

        self._num_qudits = num_qudits
        self.location = tuple(location)
        self._utry = PermutationMatrix.from_qubit_location(
            num_qudits, self.location,
        )

    def __str__(self) -> str:
        return f'PermutationGate({self.location})'

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, PermutationGate)
            and self._utry == other._utry
        )

    def __hash__(self) -> int:
        return hash(self._utry)



"""This module implements functions for translating to and from PyTKet."""
from __future__ import annotations

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import pytket

from bqskit.ir.circuit import Circuit
from bqskit.ir.lang.qasm2 import OPENQASM2Language


def pytket_to_bqskit(qc: pytket.Circuit) -> Circuit:
    """Convert PyTKet's Circuit `cc` to a BQSKit Circuit."""
    try:
        from pytket.qasm import circuit_to_qasm_str
    except ImportError as e:
        raise ImportError(
            '\n\nUnable to import pytket package.\n'
            'Please ensure that it is installed with the following command:\n'
            '\tpip install pytket\n',
        ) from e

    circuit = OPENQASM2Language().decode(circuit_to_qasm_str(qc))
    return circuit


def bqskit_to_pytket(circuit: Circuit) -> pytket.Circuit:
    """Convert a BQSKit Circuit to a PyTKet Circuit."""
    try:
        from pytket.qasm import circuit_from_qasm_str
    except ImportError as e:
        raise ImportError(
            '\n\nUnable to import pytket package.\n'
            'Please ensure that it is installed with the following command:\n'
            '\tpip install pytket\n',
        ) from e

    return circuit_from_qasm_str(OPENQASM2Language().encode(circuit))



"""This package contains composed gates."""
from __future__ import annotations

from bqskit.ir.gates.composed.controlled import ControlledGate
from bqskit.ir.gates.composed.daggergate import DaggerGate
from bqskit.ir.gates.composed.embedded import EmbeddedGate
from bqskit.ir.gates.composed.frozenparam import FrozenParameterGate
from bqskit.ir.gates.composed.tagged import TaggedGate
from bqskit.ir.gates.composed.vlg import VariableLocationGate

__all__ = [
    'ControlledGate',
    'DaggerGate',
    'EmbeddedGate',
    'FrozenParameterGate',
    'TaggedGate',
    'VariableLocationGate',
]



from __future__ import annotations

from typing import Sequence

import numpy as np
import numpy.typing as npt


class RunnerResults:
    """The result of running a Circuit through a CircuitRunner."""

    def __init__(
        self,
        num_qudits: int,
        radixes: Sequence[int],
        probs: Sequence[float],
    ) -> None:
        """Construct a RunnerResults object."""
        self.num_qudits = num_qudits
        self.radixes = radixes
        self.probs = np.array(probs)

    def get_counts(self, shots: int) -> npt.NDArray[np.int64]:
        return np.asarray(np.multiply(shots, self.probs), np.int64)

    def __str__(self) -> str:
        return str(self.probs)



"""This module implements the SubstitutePass."""
from __future__ import annotations

import itertools as it
import logging
from typing import Any
from typing import Callable

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gate import Gate
from bqskit.ir.operation import Operation
from bqskit.ir.opt.cost.functions import HilbertSchmidtResidualsGenerator
from bqskit.ir.opt.cost.generator import CostFunctionGenerator
from bqskit.utils.typing import is_real_number
_logger = logging.getLogger(__name__)


class SubstitutePass(BasePass):
    """
    The SubstitutePass class.

    The substitute pass will attempt to use instantiation to replace gates with
    other gates.
    """

    def __init__(
        self,
        collection_filter: Callable[[Operation], bool],
        gate: Gate,
        success_threshold: float = 1e-8,
        cost: CostFunctionGenerator = HilbertSchmidtResidualsGenerator(),
        instantiate_options: dict[str, Any] = {},
    ) -> None:
        """
        Construct a SubstitutePass.

        Args:
            collection_filter (Callable[[Operation], bool]):
                A predicate that determines which operations should have
                substitution attempted on them.

            gate (Gate): The gate to try substitute in.

            success_threshold (float): The distance threshold that
                determines successful termintation. Measured in cost
                described by the hilbert schmidt cost function.
                (Default: 1e-8)

            cost (CostFunction | None): The cost function that determines
                successful removal of a gate.
                (Default: HilbertSchmidtResidualsGenerator())

            instantiate_options (dict[str: Any]): Options passed directly
                to circuit.instantiate when instantiating circuit
                templates. (Default: {})
        """
        if not callable(collection_filter):
            raise TypeError(
                'Expected callable method that maps Operations to booleans for'
                ' collection_filter, got %s.' % type(self.collection_filter),
            )

        if not isinstance(gate, Gate):
            raise TypeError(f'Expected a gate, got {type(gate)}')

        if not is_real_number(success_threshold):
            raise TypeError(
                'Expected real number for success_threshold'
                ', got %s' % type(success_threshold),
            )

        if not isinstance(cost, CostFunctionGenerator):
            raise TypeError(
                'Expected cost to be a CostFunctionGenerator, got %s'
                % type(cost),
            )

        if not isinstance(instantiate_options, dict):
            raise TypeError(
                'Expected dictionary for instantiate_options, got %s.'
                % type(instantiate_options),
            )

        self.collection_filter = collection_filter
        self.gate = gate
        self.success_threshold = success_threshold
        self.cost = cost
        self.instantiate_options: dict[str, Any] = {
            'dist_tol': self.success_threshold,
        }
        self.instantiate_options.update(instantiate_options)

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        instantiate_options = self.instantiate_options.copy()
        if 'seed' not in instantiate_options:
            instantiate_options['seed'] = data.seed

        # Collect locations in circuit where self.init_gate exists
        points = [
            (cycle, op.location[0])
            for cycle, op in circuit.operations_with_cycles()
            if self.collection_filter(op)
        ]

        # Check substitutions are valid
        for point in points:
            if circuit[point].num_qudits < self.gate.num_qudits:
                raise RuntimeError(
                    f'Cannot substitute {circuit[point]}'
                    f' with {self.gate}.',
                )

        # Attempt substitution
        target = self.get_target(circuit, data)
        shift = 0
        for point in points:
            point = (point[0] - shift, point[1])
            _logger.info(f'Attempting substitution for operation at {point}.')
            _logger.info(f'Operation: {circuit[point]}')

            qudits = circuit[point].location
            locs = it.combinations(qudits, self.gate.num_qudits)
            og_cycle_count = circuit.num_cycles

            for loc in locs:
                _logger.debug(f'Trying location: {loc}')
                circuit_copy = circuit.copy()
                circuit_copy.replace_gate(point, self.gate, loc)
                circuit_copy.instantiate(target, **instantiate_options)

                if self.cost(circuit_copy, target) < self.success_threshold:
                    _logger.info('Successfully substituted operation.')
                    circuit.become(circuit_copy)
                    shift += og_cycle_count - circuit.num_cycles
                    break



"""This module implements the NotPredicate class."""
from __future__ import annotations

from typing import TYPE_CHECKING

from bqskit.passes.control.predicate import PassPredicate

if TYPE_CHECKING:
    from bqskit.compiler.passdata import PassData
    from bqskit.ir.circuit import Circuit


class NotPredicate(PassPredicate):
    """
    The NotPredicate class.

    The NotPredicate takes a predicate and always returns the opposite truth
    value.
    """

    def __init__(self, predicate: PassPredicate) -> None:
        """
        Construct a NotPredicate.

        Args:
            predicate (PassPredicate): The predicate to invert.
        """

        if not isinstance(predicate, PassPredicate):
            raise TypeError(f'Expected PassPredicate, got {type(predicate)}.')

        self.predicate = predicate

    def get_truth_value(self, circuit: Circuit, data: PassData) -> bool:
        """Call this predicate, see :class:`PassPredicate` for more info."""
        return not self.predicate(circuit, data)



"""This module implements the LBFGSMinimizer class."""
from __future__ import annotations

import logging

from bqskitrs import LBFGSMinimizerNative

from bqskit.ir.opt.minimizer import Minimizer

_logger = logging.getLogger(__name__)


class LBFGSMinimizer(LBFGSMinimizerNative, Minimizer):
    """
    The LBFGSMinimizer class.

    The LBFGSMinimizer attempts to instantiate the circuit such that the
    circuit's cost, given by a CostFunction, is minimized.
    """



"""This module implements the PauliMatrices class."""
from __future__ import annotations

import itertools as it
from typing import Iterable
from typing import Iterator
from typing import overload
from typing import Sequence
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt

from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_numeric
from bqskit.utils.typing import is_sequence

if TYPE_CHECKING:
    from bqskit.qis.unitary.unitary import RealVector


class PauliMatrices(Sequence[npt.NDArray[np.complex128]]):
    """
    Pauli group of matrices.

    A PauliMatrices object represents the entire of set of Pauli matrices for
    some number of qubits.
    """

    X = np.array(
        [
            [0, 1],
            [1, 0],
        ], dtype=np.complex128,
    )
    """The Pauli X Matrix."""

    Y = np.array(
        [
            [0, -1j],
            [1j, 0],
        ], dtype=np.complex128,
    )
    """The Pauli Y Matrix."""

    Z = np.array(
        [
            [1, 0],
            [0, -1],
        ], dtype=np.complex128,
    )
    """The Pauli Z Matrix."""

    I = np.array(
        [
            [1, 0],
            [0, 1],
        ], dtype=np.complex128,
    )
    """The Identity Matrix."""

    def __init__(self, num_qudits: int) -> None:
        """
        Construct the Pauli group for `num_qudits` number of qubits.

        Args:
            num_qudits (int): Power of the tensor product of the Pauli group.

        Raises:
            ValueError: If `num_qudits` is less than or equal to 0.
        """

        if not is_integer(num_qudits):
            raise TypeError(
                'Expected integer for num_qudits, got %s.' %
                type(num_qudits),
            )

        if num_qudits <= 0:
            raise ValueError(
                'Expected positive integer for num_qudits, got %s.' % type(
                    num_qudits,
                ),
            )

        self.num_qudits = num_qudits

        if num_qudits == 1:
            self.paulis = [
                PauliMatrices.I,
                PauliMatrices.X,
                PauliMatrices.Y,
                PauliMatrices.Z,
            ]
        else:
            self.paulis = []
            matrices = it.product(
                PauliMatrices(
                    num_qudits - 1,
                ),
                PauliMatrices(1),
            )
            for pauli_n_1, pauli_1 in matrices:
                self.paulis.append(np.kron(pauli_n_1, pauli_1))

    def __iter__(self) -> Iterator[npt.NDArray[np.complex128]]:
        return self.paulis.__iter__()

    @overload
    def __getitem__(self, index: int) -> npt.NDArray[np.complex128]:
        ...

    @overload
    def __getitem__(self, index: slice) -> list[npt.NDArray[np.complex128]]:
        ...

    def __getitem__(
        self,
        index: int | slice,
    ) -> npt.NDArray[np.complex128] | list[npt.NDArray[np.complex128]]:
        return self.paulis[index]

    def __len__(self) -> int:
        return len(self.paulis)

    @property
    def numpy(self) -> npt.NDArray[np.complex128]:
        """The NumPy array holding the pauli matrices."""
        return np.array(self.paulis)

    def __array__(
        self,
        dtype: np.typing.DTypeLike = np.complex128,
    ) -> npt.NDArray[np.complex128]:
        """Implements NumPy API for the PauliMatrices class."""
        if dtype != np.complex128:
            raise ValueError('PauliMatrices only supports Complex128 dtype.')

        return np.array(self.paulis, dtype)

    def get_projection_matrices(
            self, q_set: Iterable[int],
    ) -> list[npt.NDArray[np.complex128]]:
        """
        Return the Pauli matrices that act only on qubits in `q_set`.

        Args:
            q_set (Iterable[int]): Active qubit indices

        Returns:
            list[np.ndarray]: Pauli matrices from `self` acting only
            on qubits in `q_set`.

        Raises:
            ValueError: if `q_set` is an invalid set of qubit indices.
        """
        q_set = list(q_set)

        if not all(is_integer(q) for q in q_set):
            raise TypeError('Expected sequence of integers for qubit indices.')

        if any(q < 0 or q >= self.num_qudits for q in q_set):
            raise ValueError('Qubit indices must be in [0, n).')

        if len(q_set) != len(set(q_set)):
            raise ValueError('Qubit indices cannot have duplicates.')

        # Nth Order Pauli Matrices can be thought of base 4 number
        # I = 0, X = 1, Y = 2, Z = 3
        # XXY = 1 * 4^2 + 1 * 4^1 + 2 * 4^0 = 22 (base 10)
        # This gives the idx of XXY in paulis
        # Note we read qubit index from the left,
        # so X in XII corresponds to q = 0
        pauli_n_qubit = []
        for ps in it.product([0, 1, 2, 3], repeat=len(q_set)):
            idx = 0
            for p, q in zip(ps, q_set):
                idx += p * (4 ** (self.num_qudits - q - 1))
            pauli_n_qubit.append(self.paulis[idx])

        return pauli_n_qubit

    def dot_product(self, alpha: RealVector) -> npt.NDArray[np.complex128]:
        """
        Computes the standard dot product of `alpha` with the paulis.

        Args:
            alpha (RealVector): The pauli coefficients.

        Returns:
            np.ndarray: Sum of element-wise multiplication of `alpha`
            and `self.paulis`.

        Raises:
            ValueError: If `alpha` and `self.paulis` are incompatible.
        """

        if not is_sequence(alpha) or not all(is_numeric(a) for a in alpha):
            raise TypeError(
                'Expected a sequence of numbers, got %s.' % type(alpha),
            )

        if len(alpha) != len(self):
            raise ValueError(
                'Incorrect number of alpha values, expected %d, got %d.'
                % (len(self), len(alpha)),
            )

        return np.array(np.sum([a * s for a, s in zip(alpha, self.paulis)], 0))

    @staticmethod
    def from_string(
            pauli_string: str,
    ) -> npt.NDArray[np.complex128] | list[npt.NDArray[np.complex128]]:
        """
        Construct pauli matrices from a string description.

        Args:
            pauli_string (str): A string that describes the desired matrices.
                This is a comma-seperated list of pauli strings.
                A pauli string has the following regex pattern: [IXYZ]+

        Returns:
            np.ndarray | list[np.ndarray]: Either the single pauli matrix
            if only one is constructed, or the list of the constructed
            pauli matrices.

        Raises:
            ValueError: if `pauli_string` is invalid.
        """

        if not isinstance(pauli_string, str):
            raise TypeError(
                'Expected string for pauli_string, got %s' % type(
                    pauli_string,
                ),
            )

        pauli_strings = [
            string.strip().upper()
            for string in pauli_string.split(',')
            if len(string.strip()) > 0
        ]

        pauli_matrices = []
        idx_dict = {'I': 0, 'X': 1, 'Y': 2, 'Z': 3}
        mat_dict = {
            'I': PauliMatrices.I,
            'X': PauliMatrices.X,
            'Y': PauliMatrices.Y,
            'Z': PauliMatrices.Z,
        }

        for pauli_string in pauli_strings:
            if not all(char in 'IXYZ' for char in pauli_string):
                raise ValueError('Invalid pauli string.')

            if len(pauli_string) <= 6:
                idx = 0
                for char in pauli_string:
                    idx *= 4
                    idx += idx_dict[char]
                pauli_matrices.append(PauliMatrices(len(pauli_string))[idx])
            else:
                acm = mat_dict[pauli_string[0]]
                for char in pauli_string[1:]:
                    acm = np.kron(acm, mat_dict[char])
                pauli_matrices.append(acm)

        if len(pauli_matrices) == 1:
            return pauli_matrices[0]

        return pauli_matrices



"""This module implements the PassData class."""
from __future__ import annotations

import copy
import itertools as it
from typing import Any
from typing import Iterator
from typing import MutableMapping
from typing import Sequence

from bqskit.compiler.gateset import GateSet
from bqskit.compiler.machine import MachineModel
from bqskit.ir.circuit import Circuit
from bqskit.qis.graph import CouplingGraph
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_real_number
from bqskit.utils.typing import is_sequence


class PassData(MutableMapping[str, Any]):
    """
    A dictionary wrapper shared between all passes in a compilation run.

    This class reserves certain keywords and supplies default initial values for
    them. Other than that, it behaves very similar to a normal dictionary object
    mapping `str` keys to any type of value.
    """

    _reserved_keys = [
        'target',
        'model',
        'placement',
        'error',
        'seed',
        'machine_model',
        'initial_mapping',
        'final_mapping',
    ]

    def __init__(self, circuit: Circuit) -> None:
        """Initialize a PassData object from `circuit`."""
        self._target: Circuit | StateVector | UnitaryMatrix | StateSystem
        if circuit.num_qudits <= 8:
            try:
                self._target = circuit.get_unitary()
            except RuntimeError:
                self._target = circuit
        else:
            self._target = circuit  # Lazy evaluation

        self._error = 0.0
        self._model = MachineModel(circuit.num_qudits)
        self._placement = list(range(circuit.num_qudits))
        self._initial_mapping = list(range(circuit.num_qudits))
        self._final_mapping = list(range(circuit.num_qudits))
        self._data: dict[str, Any] = {}
        self._seed: int | None = None

    @property
    def target(self) -> StateVector | UnitaryMatrix | StateSystem:
        """Return the current target unitary or state."""
        if isinstance(self._target, Circuit):
            self._target = self._target.get_unitary()

        return self._target

    @target.setter
    def target(self, _val: StateVector | UnitaryMatrix | StateSystem) -> None:
        if not isinstance(_val, (StateVector, UnitaryMatrix, StateSystem)):
            raise TypeError(
                f'Cannot assign type {type(_val)} to target.'
                ' Expected either a StateVector, StateSystem,'
                ' or UnitaryMatrix.',
            )
        if len(self.placement) != _val.num_qudits:
            self.placement = list(range(_val.num_qudits))
        self._target = _val

    @property
    def error(self) -> float:
        """Return the current target unitary or state."""
        return self._error

    @error.setter
    def error(self, _val: float) -> None:
        if not is_real_number(_val):
            raise TypeError(
                f'Cannot assign type {type(_val)} to error.'
                ' Expected a real number.',
            )

        self._error = _val

    @property
    def model(self) -> MachineModel:
        """Return the current target MachineModel."""
        return self._model

    @model.setter
    def model(self, _val: MachineModel) -> None:
        if not isinstance(_val, MachineModel):
            raise TypeError(
                f'Cannot set model to {type(_val)}.'
                ' Expected a MachineModel.',
            )

        self._model = _val

    @property
    def gate_set(self) -> GateSet:
        """Return the current target MachineModel's GateSet."""
        return self._model.gate_set

    @gate_set.setter
    def gate_set(self, _val: GateSet) -> None:
        if not isinstance(_val, GateSet):
            raise TypeError(
                f'Cannot set gate_set to {type(_val)}.'
                ' Expected a GateSet.',
            )

        self._model.gate_set = _val

    @property
    def placement(self) -> list[int]:
        """Return the current placement of circuit qudits on model qudits."""
        return self._placement

    @placement.setter
    def placement(self, _val: Sequence[int]) -> None:
        if not is_sequence(_val):
            raise TypeError(
                f'Cannot set placement to {type(_val)}.'
                ' Expected a sequence of integers.',
            )

        if not all(is_integer(x) for x in _val):
            raise TypeError(
                'Cannot set placement. Expected a sequence of integers.',
            )

        self._placement = list(int(x) for x in _val)

    @property
    def initial_mapping(self) -> list[int]:
        """
        Return the initial mapping of logical to physical qudits.

        This always maps how the logical qudits from the original circuit start
        on the physical qudits of the current circuit.
        """
        return self._initial_mapping

    @initial_mapping.setter
    def initial_mapping(self, _val: Sequence[int]) -> None:
        if not is_sequence(_val):
            raise TypeError(
                f'Cannot set initial_mapping to {type(_val)}.'
                ' Expected a sequence of integers.',
            )

        if not all(is_integer(x) for x in _val):
            raise TypeError(
                'Cannot set initial_mapping. Expected a sequence of integers.',
            )

        self._initial_mapping = list(int(x) for x in _val)

    @property
    def final_mapping(self) -> list[int]:
        """
        Return the final mapping of logical to physical qudits.

        This always maps how the logical qudits from the original circuit end on
        the physical qudits of the current circuit.
        """
        return self._final_mapping

    @final_mapping.setter
    def final_mapping(self, _val: Sequence[int]) -> None:
        if not is_sequence(_val):
            raise TypeError(
                f'Cannot set final_mapping to {type(_val)}.'
                ' Expected a sequence of integers.',
            )

        if not all(is_integer(x) for x in _val):
            raise TypeError(
                'Cannot set final_mapping. Expected a sequence of integers.',
            )

        self._final_mapping = list(int(x) for x in _val)

    @property
    def seed(self) -> int | None:
        """Return the pass's seed."""
        return self._seed

    @seed.setter
    def seed(self, _val: int | None) -> None:
        if _val is not None and not is_integer(_val):
            raise TypeError(
                f'Cannot set seed to {type(_val)}.'
                ' Expected an integer or none.',
            )
        self._seed = _val

    @property
    def connectivity(self) -> CouplingGraph:
        """Retrieve the physical connectivity of the circuit qudits."""
        return self.model.coupling_graph.get_subgraph(self.placement)

    def __getitem__(self, _key: str) -> Any:
        """Retrieve the value associated with `_key` from the pass data."""
        if _key in self._reserved_keys:
            if _key == 'machine_model':
                _key = 'model'
            return self.__getattribute__(_key)

        return self._data.__getitem__(_key)

    def __setitem__(self, _key: str, _val: Any) -> None:
        """Update the value associated with `_key` in the pass data."""
        if _key in self._reserved_keys:
            if _key == 'machine_model':
                _key = 'model'
            return self.__setattr__(_key, _val)

        return self._data.__setitem__(_key, _val)

    def __delitem__(self, _key: str) -> None:
        """Delete the key-value pair associated with `_key`."""
        if _key in self._reserved_keys:
            raise RuntimeError(f'Cannot delete {_key} from data.')

        return self._data.__delitem__(_key)

    def __iter__(self) -> Iterator[str]:
        """Return an iterator over all keys in the pass data."""
        return it.chain(self._reserved_keys.__iter__(), self._data.__iter__())

    def __len__(self) -> int:
        """Return the number of key-value pairs in the pass data."""
        return self._data.__len__() + len(self._reserved_keys)

    def __contains__(self, _o: object) -> bool:
        """Return true if `_o` is a key in the pass data."""
        in_resv = self._reserved_keys.__contains__(_o)
        in_data = self._data.__contains__(_o)
        return in_resv or in_data

    def copy(self) -> PassData:
        """Returns a deep copy of the data."""
        return copy.deepcopy(self)

    def become(self, other: PassData, deepcopy: bool = False) -> None:
        """Become a copy of `other`."""
        if deepcopy:
            self._target = copy.deepcopy(other._target)
            self._error = copy.deepcopy(other._error)
            self._model = copy.deepcopy(other._model)
            self._placement = copy.deepcopy(other._placement)
            self._data = copy.deepcopy(other._data)
            self._seed = copy.deepcopy(other._seed)
        else:
            self._target = copy.copy(other._target)
            self._error = copy.copy(other._error)
            self._model = copy.copy(other._model)
            self._placement = copy.copy(other._placement)
            self._data = copy.copy(other._data)
            self._seed = copy.copy(other._seed)

    def update_error_mul(self, error: float) -> None:
        """Update the error multiplicatively."""
        self.error = (1 - ((1 - self.error) * (1 - error)))



"""This module implements the RuntimeTask class."""
from __future__ import annotations

import inspect
import logging
from typing import Any
from typing import Coroutine

from bqskit.runtime.address import RuntimeAddress


class RuntimeTask:
    """
    BQSKit Runtime's internal task structure.

    The task exists in two phases:
        1. In the initial phase, the task is a context-less function
        and arguments grouped with a return address. It is expected
        when the task is finished executing, to ship the result to
        the return address.

        2. When the task has been scheduled to run on a worker, it is
        initialized into a coroutine that can be executed in steps.
    """

    task_counter = 0
    """Per-process task id counter."""

    def __init__(
        self,
        fnargs: tuple[Any, Any, Any],
        return_address: RuntimeAddress,
        comp_task_id: int,
        breadcrumbs: tuple[RuntimeAddress, ...],
        logging_level: int | None = None,
        max_logging_depth: int = -1,
    ) -> None:
        """Create the task with a new id and return address."""
        RuntimeTask.task_counter += 1
        self.task_id = RuntimeTask.task_counter

        self.fnargs = fnargs
        """Tuple of function pointer, arguments, and keyword arguments."""

        self.return_address = return_address
        """Where the result of this task should be sent."""

        self.logging_level = logging_level
        """Logs with levels >= to this get emitted, if None always emit."""

        self.comp_task_id = comp_task_id
        """The mailbox id of the this task's root task."""

        self.breadcrumbs = breadcrumbs
        """All of this task's parent tasks' addresses in order."""

        self.max_logging_depth = max_logging_depth
        """Logs are not emitted for tasks with this many or more parents."""

        self.coro: Coroutine[Any, Any, Any] | None = None
        """The coroutine containing this tasks code."""

        # self.send: Any = None
        # """A register that both the coroutine and task have access to."""

        self.desired_box_id: int | None = None
        """When waiting on a mailbox, this stores that mailbox's id."""

        self.owned_mailboxes: list[int] = []
        """The mailbox ids that this task owns."""

        self.wake_on_next: bool = False
        """Set to true if this task should wake immediately on a result."""

    def step(self, send_val: Any = None) -> Any:
        """Execute one step of the task."""
        if self.coro is None:
            raise RuntimeError('Task has not been initialized.')

        # Reset previously set await flags
        self.wake_on_next = False
        self.desired_box_id = None

        # Set logging level
        old_level = logging.getLogger().getEffectiveLevel()
        if (
            self.max_logging_depth < 0
            or len(self.breadcrumbs) <= self.max_logging_depth
        ):
            logging.getLogger().setLevel(0)

        # Execute a task step
        to_return = self.coro.send(send_val)

        # Reset logging
        logging.getLogger().setLevel(old_level)

        return to_return

    def start(self) -> None:
        """Initialize the task."""
        self.coro = self.run()

    async def run(self) -> Any:
        """Task coroutine wrapper."""
        if inspect.iscoroutinefunction(self.fnargs[0]):
            return await self.fnargs[0](*self.fnargs[1], **self.fnargs[2])
        return self.fnargs[0](*self.fnargs[1], **self.fnargs[2])

    def is_descendant_of(self, addr: RuntimeAddress) -> bool:
        """Return true if `addr` identifies a parent (or this) task."""
        return addr == self.return_address or addr in self.breadcrumbs



"""This module implements the WidthPredicate class."""
from __future__ import annotations

from typing import TYPE_CHECKING

from bqskit.passes.control.predicate import PassPredicate
from bqskit.utils.typing import is_integer

if TYPE_CHECKING:
    from bqskit.compiler.passdata import PassData
    from bqskit.ir.circuit import Circuit


class WidthPredicate(PassPredicate):
    """
    The WidthPredicate class.

    The WidthPredicate class returns True if the circuit's width (number of
    qudits) is less than a specified number.
    """

    def __init__(self, width: int) -> None:
        """
        Construct a WidthPredicate.

        Args:
            width (int): Return true if the circuit's width is less than this.
        """

        if not is_integer(width):
            raise TypeError(f'Expected int, got {type(width)}')

        self.width = int(width)

    def get_truth_value(self, circuit: Circuit, data: PassData) -> bool:
        """Call this predicate, see :class:`PassPredicate` for more info."""
        return circuit.num_qudits < self.width



"""This module implements the CouplingGraph class."""
from __future__ import annotations

import copy
import itertools as it
import logging
from random import shuffle
from typing import Any
from typing import cast
from typing import Collection
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Tuple
from typing import TYPE_CHECKING
from typing import Union

import numpy as np

if TYPE_CHECKING:
    from typing_extensions import TypeGuard

from bqskit.ir.location import CircuitLocation
from bqskit.ir.location import CircuitLocationLike
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_iterable

_logger = logging.getLogger(__name__)


class CouplingGraph(Collection[Tuple[int, int]]):
    """A graph representing connections in a qudit topology."""

    def __init__(
        self,
        graph: Iterable[tuple[int, int]],
        num_qudits: int | None = None,
    ) -> None:
        if isinstance(graph, CouplingGraph):
            self.num_qudits: int = graph.num_qudits
            self._edges: set[tuple[int, int]] = graph._edges
            self._adj: list[set[int]] = graph._adj
            return

        if not CouplingGraph.is_valid_coupling_graph(graph):
            raise TypeError('Invalid coupling graph.')

        self._edges = {g if g[0] <= g[1] else (g[1], g[0]) for g in graph}

        calced_num_qudits = 0
        for q1, q2 in self._edges:
            calced_num_qudits = max(calced_num_qudits, max(q1, q2))
        calced_num_qudits += 1

        if num_qudits is None:
            self.num_qudits = calced_num_qudits
        elif calced_num_qudits > num_qudits:
            raise ValueError('Edges between invalid qudits.')
        else:
            self.num_qudits = num_qudits

        self._adj = [set() for _ in range(self.num_qudits)]
        for q1, q2 in self._edges:
            self._adj[q1].add(q2)
            self._adj[q2].add(q1)

        self._mat = [
            [np.inf for _ in range(self.num_qudits)]
            for _ in range(self.num_qudits)
        ]
        for q1, q2 in self._edges:
            self._mat[q1][q2] = 1
            self._mat[q2][q1] = 1

    def is_fully_connected(self) -> bool:
        """Return true if the graph is fully connected."""
        frontier: set[int] = {0}
        qudits_seen: set[int] = set()

        while len(frontier) > 0:
            expanded_qudits = set()

            for qudit in frontier:
                expanded_qudits.add(qudit)
                expanded_qudits.update(self._adj[qudit])

            frontier = expanded_qudits - qudits_seen
            qudits_seen.update(expanded_qudits)

            if len(qudits_seen) == self.num_qudits:
                return True

        return False

    def get_neighbors_of(self, qudit: int) -> list[int]:
        """Return the qudits adjacent to `qudit`."""
        return list(self._adj[qudit])

    def __contains__(self, __o: object) -> bool:
        return self._edges.__contains__(__o)

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, CouplingGraph):
            return False

        if self.num_qudits != __o.num_qudits:
            return False

        if self._mat != __o._mat:
            return False

        return True

    def __iter__(self) -> Iterator[tuple[int, int]]:
        return self._edges.__iter__()

    def __hash__(self) -> int:
        return hash((self.num_qudits, tuple(self._edges)))

    def __len__(self) -> int:
        return self._edges.__len__()

    def __str__(self) -> str:
        return 'CouplingGraph(' + self._edges.__str__() + ')'

    def __repr__(self) -> str:
        return self._edges.__repr__()

    def get_qudit_degrees(self) -> list[int]:
        return [len(l) for l in self._adj]

    def all_pairs_shortest_path(self) -> list[list[int]]:
        """
        Calculate all pairs shortest path matrix using Floyd-Warshall.

        Returns:
            D (list[list[int]]): D[i][j] is the length of the shortest
                path from i to j.
        """
        D = copy.deepcopy(self._mat)
        for k in range(self.num_qudits):
            for i in range(self.num_qudits):
                for j in range(self.num_qudits):
                    D[i][j] = min(D[i][j], D[i][k] + D[k][j])
        return cast(List[List[int]], D)

    def get_shortest_path_tree(self, source: int) -> list[tuple[int, ...]]:
        """Return shortest path from `source` to every node in `self`."""
        # Dijkstra's algorithm to build shortest-path tree
        unvisited_qudits = set(range(self.num_qudits))
        distances = {i: np.inf for i in range(self.num_qudits)}
        paths: list[tuple[int, ...]] = [tuple() for i in range(self.num_qudits)]
        distances[source] = 0
        paths[source] = (source,)

        while len(unvisited_qudits) > 0:
            # Pick next unvisited qudit with shortest distance
            unvisited_distances = [
                (x, y) for x, y in distances.items() if x in unvisited_qudits
            ]
            unvisited_distances.sort(key=lambda x: x[1])
            current_qudit = unvisited_distances[0][0]

            if distances[current_qudit] == np.inf:
                raise RuntimeError(f'No path found to qudit: {current_qudit}.')

            neighbors = self.get_neighbors_of(current_qudit)
            unvisted_neighbors = unvisited_qudits.intersection(neighbors)

            for other_qudit in unvisted_neighbors:
                if distances[current_qudit] + 1 < distances[other_qudit]:
                    distances[other_qudit] = distances[current_qudit] + 1
                    paths[other_qudit] = paths[current_qudit] + (other_qudit,)

            unvisited_qudits.remove(current_qudit)

        return [paths[i] for i in range(self.num_qudits)]

    def get_subgraph(
        self,
        location: CircuitLocationLike,
        renumbering: dict[int, int] | None = None,
    ) -> CouplingGraph:
        """Returns the sub-coupling-graph with qudits in `location`."""
        if not CircuitLocation.is_location(location, self.num_qudits):
            raise TypeError('Invalid location.')

        location = CircuitLocation(location)
        if renumbering is None:
            renumbering = {q: i for i, q in enumerate(location)}

        subgraph = []
        location_set = {loc for loc in location}
        for q_i in location:
            for q_i_neighbor in location_set.intersection(self._adj[q_i]):
                subgraph.append((renumbering[q_i], renumbering[q_i_neighbor]))
        return CouplingGraph(subgraph, len(location))

    def get_subgraphs_of_size(self, size: int) -> list[CircuitLocation]:
        """
        Find all sets of indices that form connected subgraphs on their own.

        Each location describes a valid spot for a `size`-sized gate,
        so the number of qudit_indices in each location is `size`.
        A location is only included if each pair of qudits is directly
        connected or connected through other qudits in the location.

        Args:
            size (int): The size of each location in the final list.

        Returns:
            list[CircuitLocation]: The locations compliant with the machine.

        Raises:
            ValueError: If `size` is nonpositive or too large.

        Notes:
            Does a breadth first search on all pairs of qudits, keeps paths
            that have length equal to size. Note that the coupling map
            is assumed to be undirected.
        """
        if not is_integer(size):
            raise TypeError(
                f'Expected integer for block_size, got {type(size)}',
            )

        if size > self.num_qudits:
            raise ValueError(
                'The block_size is too large; '
                f'graph size is {self.num_qudits}, got {size}.',
            )

        if size <= 0:
            raise ValueError(f'Nonpositive size; got {size}.')

        locations: set[CircuitLocation] = set()

        for qudit in range(self.num_qudits):
            # Get every valid set containing `qudit` with size == size
            self._location_search(locations, set(), qudit, size)

        return list(locations)

    def _location_search(
        self,
        locations: set[CircuitLocation],
        path: set[int],
        vertex: int,
        limit: int,
    ) -> None:
        """
        Add paths with length equal to limit to the `locations` set.

        Args:
            locations (set[CircuitLocation]): A list that contains all paths
                found so far of length equal to `limit`.

            path (set[int]): The qudit vertices currently included in
                the path.

            vertex (int): The vertex in the graph currently being examined.

            limit (int): The desired length of paths in the `locations`
                list.
        """
        if vertex in path:
            return

        curr_path = path.copy()
        curr_path.add(vertex)

        if len(curr_path) == limit:
            locations.add(CircuitLocation(list(curr_path)))
            return

        frontier: set[int] = {
            qudit
            for node in curr_path
            for qudit in self._adj[node]
            if qudit not in curr_path
        }

        for neighbor in frontier:
            self._location_search(locations, curr_path, neighbor, limit)

    @staticmethod
    def is_valid_coupling_graph(
        coupling_graph: Any,
        num_qudits: int | None = None,
    ) -> TypeGuard[CouplingGraphLike]:
        """
        Return true if the coupling graph is valid.

        Args:
            coupling_graph (Any): The coupling graph to check.

            num_qudits (int | None): The total number of qudits. All qudits
                should be less than this. If None, don't check.

        Returns:
            bool: Valid or not
        """

        if not is_iterable(coupling_graph):
            _logger.debug('Coupling graph is not iterable.')
            return False

        if len(list(coupling_graph)) == 0:
            return True

        if not all(isinstance(pair, tuple) for pair in coupling_graph):
            _logger.debug('Coupling graph is not a sequence of tuples.')
            return False

        if not all([len(pair) == 2 for pair in coupling_graph]):
            _logger.debug('Coupling graph is not a sequence of pairs.')
            return False

        if num_qudits is not None:
            if not (is_integer(num_qudits) and num_qudits > 0):
                _logger.debug('Invalid num_qudits in coupling graph check.')
                return False

            if not all(
                qudit < num_qudits
                for pair in coupling_graph
                for qudit in pair
            ):
                _logger.debug('Coupling graph has invalid qudits.')
                return False

        if not all([
            len(pair) == len(set(pair))
            for pair in coupling_graph
        ]):
            _logger.debug('Coupling graph has an invalid pair.')
            return False

        return True

    @staticmethod
    def all_to_all(num_qudits: int) -> CouplingGraph:
        """Return a coupling graph with all qudits connected."""
        return CouplingGraph(set(it.combinations(range(num_qudits), 2)))

    @staticmethod
    def linear(num_qudits: int) -> CouplingGraph:
        """Return a coupling graph with nearest-neighbor connectivity."""
        return CouplingGraph([(x, x + 1) for x in range(num_qudits - 1)])

    @staticmethod
    def ring(num_qudits: int) -> CouplingGraph:
        """Return a coupling graph with ring connectivity."""
        ext = [(0, num_qudits - 1)]
        return CouplingGraph([(x, x + 1) for x in range(num_qudits - 1)] + ext)

    @staticmethod
    def star(num_qudits: int) -> CouplingGraph:
        """Return a coupling graph with one qudit connected to all else."""
        return CouplingGraph([(0, x) for x in range(1, num_qudits)])

    @staticmethod
    def grid(num_rows: int, num_cols: int) -> CouplingGraph:
        """Return a coupling graph with a grid of qubits."""
        num_qudits = num_rows * num_cols
        edges = set()

        for i in range(num_qudits):
            if i % num_cols != num_cols - 1:
                edges.add((i, i + 1))

            if i < (num_rows - 1) * num_cols:
                edges.add((i, i + num_cols))

        return CouplingGraph(edges)

    def maximal_matching(
        self,
        edges_to_ignore: list[tuple[int, int]] = [],
        randomize: bool = False,
    ) -> list[tuple[int, int]]:
        """
        Generate a random graph matching for the coupling graph. Edges in the
        `ignored_edges` list will not be included.

        Arguments:
            edges_to_ignore (list[tuple[int]]): Edges not included in the
                matching. (Default: [])

            randomize (bool): Shuffle edges to create random matchings.
                (Default: False)

        Returns:
            matching (list[tuple[int]]): A maximal list of edges that share
                no common verticies.
        """
        matching: set[tuple[int, int]] = set()
        vertices: set[int] = set()

        edge_list = [
            (u, v) for (u, v) in self._edges if (u, v) not in edges_to_ignore
            and (v, u) not in edges_to_ignore
        ]
        if randomize:
            shuffle(edge_list)

        for edge in edge_list:
            u, v = edge
            if u not in vertices and v not in vertices and u != v:
                matching.add(edge)
                vertices.update(edge)
        return list(matching)

    def is_fully_connected_without(self, qudit: int) -> bool:
        """Return true if the graph is fully connected without `qudit`."""
        frontier = {0} if qudit != 0 else {1}
        qudits_seen = {0} if qudit != 0 else {1}

        while len(frontier) > 0:
            expanded_qudits = set()

            for q in frontier:
                if q != qudit:
                    neighbors = set(self.get_neighbors_of(q)) - {qudit}
                    expanded_qudits.update(neighbors)

            frontier = expanded_qudits - qudits_seen
            qudits_seen.update(expanded_qudits)

            if len(qudits_seen) == self.num_qudits - 1:
                return True

        return False

    def get_rooted_minimum_span(self, root: int) -> list[tuple[int, int]]:
        """
        Connect `root` to every other node in the graph.

        Args:
            root (int): The qudit to start from.

        Return:
            (list[tuple[int, int]]): A list of minimal edges that
                connects 'root' to every other qudit.
        """
        # Construct Minimum Spanning Tree using breadth-first search
        mst = []
        seen = {root}
        frontier = [root]

        while len(frontier) > 0 and len(seen) < self.num_qudits:
            qudit = frontier.pop(0)
            neighbors = set(self.get_neighbors_of(qudit))
            unseen_neighbors = neighbors - seen
            frontier.extend(unseen_neighbors)
            seen.update(neighbors)
            for neighbor in unseen_neighbors:
                mst.append((qudit, neighbor))

        mst = CouplingGraph(mst)

        # Traverse MST depth-first to produce pairs of interacting qudits
        interactions: list[tuple[int, int]] = []
        frontier: list[tuple[int, tuple[int, int] | None]] = [(root, None)]

        while len(frontier) > 0:
            qudit, interaction = frontier.pop(0)
            if interaction is not None:
                interactions.append(interaction)

            for neighbor in mst.get_neighbors_of(qudit):
                if interaction is None or neighbor != interaction[0]:
                    frontier.insert(0, (neighbor, (qudit, neighbor)))

        return interactions

    def get_induced_subgraph(
        self,
        location: CircuitLocationLike,
    ) -> list[tuple[int, int]]:
        """
        Return the edges induced by the vertices specified in `location`.

        Arguments:
            location (CircuitLocationLike): A list of qubits or vertices in the
                CouplingGraph. The edges connecting each of these vertices
                together will be extracted.

        Returns:
            (list[tuple[int,int]]): The list of edges connecting vertices in
                `location`.
        """
        if not isinstance(location, CircuitLocation):
            location = CircuitLocation(location)

        if len(location) < 2:
            raise ValueError('Invalid location size.')

        subgraph = []
        for q1, q2 in it.combinations(location, 2):
            if (q1, q2) in self._edges or (q2, q1) in self._edges:
                subgraph.append((min([q1, q2]), max([q1, q2])))
        return subgraph

    def relabel_subgraph(
        graph: CouplingGraphLike,
        relabeling: dict[int, int] | None = None,
    ) -> CouplingGraph:
        """
        Renumber the vertices in `graph` according to the optionally provided
        `relabeling` dictionary, or relabel so that the vertices are in
        renumbered in least to greatest order and in the set {0,...,|V|-1}.

        Arguments:
            graph (CouplingGraphLike): A collection of edges representing a
                graph.

            relabeling (dict[int,int]|None): An optional dictionary specifying
                current labels as keys and desired labels as values.
                (Default: None)

        Returns:
            (CouplingGraph): The relabeled CouplingGraph.
        """
        if relabeling is None:
            vertices = set()
            for q1, q2 in graph:
                vertices.add(q1)
                vertices.add(q2)
            renum = {q: i for i, q in enumerate(vertices)}
        else:
            renum = relabeling

        edges = [
            (min([renum[q1], renum[q2]]), max([renum[q1], renum[q2]]))
            for q1, q2 in graph
        ]
        return CouplingGraph(edges)

    def is_embedded_in(self, graph: CouplingGraph) -> bool:
        """
        Check if this CouplingGraph is embedded within `graph`.

        Arguments:
            graph (CouplingGraph): A graph that may have a subgraph isomorphic
                to self.

        Returns:
            (bool): True if this graph is isomorphic to a subgraph of `graph`,
                and False if not.

        Note:
            The algorithm used is naive and relies on the fact that `graph` and
            `self` are similar in size and sparse.

        Todo:
            Implement the V2 algorithm for subgraph isomorphism.
        """
        if type(graph) is not CouplingGraph:
            raise ValueError('Provided `graph` is not of type CouplingGraph.')
        # A larger graph cannot be embedded in a smaller graph
        if self.num_qudits > graph.num_qudits:
            return False

        degs = {q: len(self._adj[q]) for q in range(self.num_qudits)}
        other_degs = {q: len(graph._adj[q]) for q in range(graph.num_qudits)}

        # Candidates are vertices in the other graph that have degree >= a
        # given vertex in self.
        candidate_labels: dict[int, list[int]] = {
            q: [] for q in range(self.num_qudits)
        }

        for q1 in range(self.num_qudits):
            for q2 in range(graph.num_qudits):
                if degs[q1] <= other_degs[q2]:
                    candidate_labels[q1].append(q2)

        # Each vertex must have at least one candidate vertex
        if any([len(cands) == 0 for cands in candidate_labels.values()]):
            return False

        # Check if the current renumbering works
        for renumbering in it.permutations(
            range(graph.num_qudits), self.num_qudits,
        ):
            renum = {q: renumbering[q] for q in range(self.num_qudits)}
            if all([
                (min([renum[u], renum[v]]), max([renum[u], renum[v]]))
                in graph for u, v in self._edges
            ]):
                return True
        return False


CouplingGraphLike = Union[Iterable[Tuple[int, int]], CouplingGraph]



"""This module implements the CostFunction base classes."""
from __future__ import annotations

import abc

from bqskit.qis.unitary.unitary import RealVector
from bqskit.utils.typing import is_real_number
from bqskit.utils.typing import is_sequence


class CostFunction(abc.ABC):
    """
    The CostFunction base class.

    A CostFunction is a map from a vector of real numbers to a real number. This
    output represents cost and minimizers will attempt to reduce it. The input
    should represent circuit parameters.
    """

    @abc.abstractmethod
    def get_cost(self, params: RealVector) -> float:
        """Return the cost value given the input parameters."""

    def __call__(self, params: RealVector) -> float:
        """Return the cost value given the input parameters."""

        if not is_sequence(params):
            raise TypeError(
                'Expected sequence for params, got %s.' % type(params),
            )

        if not all(is_real_number(param) for param in params):
            raise TypeError(
                'Expected sequence of floats for params.',
            )

        return self.get_cost(params)



"""This package implements passes that perform qudit assignment."""
from __future__ import annotations

from bqskit.passes.mapping.apply import ApplyPlacement
from bqskit.passes.mapping.embed import EmbedAllPermutationsPass
from bqskit.passes.mapping.layout.pam import PAMLayoutPass
from bqskit.passes.mapping.layout.sabre import GeneralizedSabreLayoutPass
from bqskit.passes.mapping.placement.greedy import GreedyPlacementPass
from bqskit.passes.mapping.placement.trivial import TrivialPlacementPass
from bqskit.passes.mapping.routing.pam import PAMRoutingPass
from bqskit.passes.mapping.routing.sabre import GeneralizedSabreRoutingPass
from bqskit.passes.mapping.setmodel import ExtractModelConnectivityPass
from bqskit.passes.mapping.setmodel import RestoreModelConnevtivityPass
from bqskit.passes.mapping.setmodel import SetModelPass
from bqskit.passes.mapping.topology import SubtopologySelectionPass
from bqskit.passes.mapping.verify import CalculatePAMErrorsPass
from bqskit.passes.mapping.verify import PAMVerificationSequence
from bqskit.passes.mapping.verify import TagPAMBlockDataPass
from bqskit.passes.mapping.verify import UnTagPAMBlockDataPass

__all__ = [
    'GeneralizedSabreLayoutPass',
    'GreedyPlacementPass',
    'TrivialPlacementPass',
    'GeneralizedSabreRoutingPass',
    'SetModelPass',
    'ApplyPlacement',
    'PAMLayoutPass',
    'PAMRoutingPass',
    'EmbedAllPermutationsPass',
    'SubtopologySelectionPass',
    'ExtractModelConnectivityPass',
    'RestoreModelConnevtivityPass',
    'TagPAMBlockDataPass',
    'CalculatePAMErrorsPass',
    'UnTagPAMBlockDataPass',
    'PAMVerificationSequence',
]



"""This module implements the StructureAnalysisPass."""
from __future__ import annotations

import logging

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.circuit import CircuitGate
from bqskit.ir.structure import CircuitStructure

_logger = logging.getLogger(__name__)


class StructureAnalysisPass(BasePass):
    """
    A pass that catalogs circuit structures used in a partitioned circuit.

    If a circuit contains hierarchically defined gates, these hierarchies
    are discarded to check gate level equivalence between structures.

    Example:
        The left shows a hierarchically defined gate sequence. The
        right shows the same gate sequence but it is specified at the
        gate level. These two structures are considered equivalent.
           ----------    ----------
        --|-cx-rz-cx-|--|-cx-rz-cx-|--    --cx-rz-cx-cx-rz-cx--
          | |     |  |  | |     |  |   ==   |     |  |     |
        --|-cx-rz-cx-|--|-cx-rz-cx-|--    --cx-rz-cx-cx-rz-cx--
           ----------    ----------
    """

    def __init__(self) -> None:
        """Construct a StructurePass."""

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""

        structures_seen: dict[CircuitStructure, int] = {}

        for block in circuit:
            if isinstance(block.gate, CircuitGate):
                subcirc = block.gate._circuit
                # Structure depends on the gate level, call unfold_all
                subcirc.unfold_all()
                structure = CircuitStructure(subcirc)
                if structure not in structures_seen:
                    structures_seen[structure] = 1
                else:
                    structures_seen[structure] += 1

        if 'structures' in data:
            _logger.warning('Overriding `structures` field in PassData.')
        data['structures'] = structures_seen
